#============================================================================
#
# Program: libcli-sshnet.pl
# Programmer: Remo Rickli, dcr
#
# -> Net::Telnet/Net::SSH::Perl based Functions <-
#
# Needs quite some perl libs, but works well on OBSD4.  Rather limited at this stage:
#
# Ubuntu install hint
#$ wget http://search.cpan.org/CPAN/authors/id/D/DB/DBROBINS/Net-SSH-Perl-1.30.tar.gz
#$ tar zxvf Net-SSH-Perl-1.30.tar.gz
#$ cd Net-SSH-Perl-1.30
#$ perl Makefile.PL && make
#sudo checkinstall -D --pkgname="Marc-Net-SSH-Perl" --pkgversion="1.30" make install
#
# Foundry only tested with simple telnet pw/en configs as Radius enable asks for username again!
# HP Procurve only works with "console terminal NONE", maybe they can add vt100 without 
# ESC sequences and disable "anykey" at login?
# Nortel works best with "banner disabled"
#
# PrepDev() tries to find credentials on new devices (or if cliport set to 0).
# After that BridgeFwd() reads MAC table on supported switches if PrepDev()
# confirmed support (only IOS at the time).
# Config() fetches the configuration and stores the "interesting" part.
# All functions use the universal EnableDev() to get into enable mode.
# The $obj->get is used to avoid problems with ^M and Escape sequences...
#
# If you want to debug uncomment input.log and output.log and open extra terminals with tail -f
# on them to see what's happening right away.
#============================================================================
package cli;
use Net::Telnet;

use vars qw($sshnetok %cmd);

# Commands and prompts. Adjust if you encounter problems.
#============================================================================
$cmd{'IOS'}{'prom'} = '/(?m:^[\w.,-\@\~\/\\\+\!=:;{}\[\]]+\s?(?:\(config[^\)]*\))?\s?[#>]\s*$)/';
$cmd{'IOS'}{'conf'} = 'show run';
$cmd{'IOS'}{'page'} = 'terminal length 0';
$cmd{'IOS'}{'macd'} = 'sh mac-address-table dyn';
$cmd{'IOS'}{'macs'} = 'sh port-security addr';							# Thanks Duane Walker 7/2007
$cmd{'IOS'}{'paus'} = '0.3';
$cmd{'IOS-wl'}{'prom'} = $cmd{'IOS'}{'prom'};
$cmd{'IOS-wl'}{'conf'} = 'show run';
$cmd{'IOS-wl'}{'page'} = 'terminal length 0';
$cmd{'IOS-wl'}{'macd'} = 'sh bridge | exclude \*\*\*';
$cmd{'IOS-wl'}{'paus'} = '0.3';
$cmd{'IOS-fw'}{'prom'} = '/.+?[#>]\s?$/';							# ^M at beginning of line!
$cmd{'IOS-fw'}{'conf'} = 'show run';
$cmd{'IOS-fw'}{'page'} = 'no pager';								# PIX 6.3(3) requires this
#$cmd{'IOS-fw'}{'page'} = 'no terminal pager';							# PIX 8.0.3 seems to require this
$cmd{'IOS-fw'}{'paus'} = '0.5';
$cmd{'IOS-fv'}{'prom'} = $cmd{'IOS'}{'prom'};
$cmd{'IOS-fv'}{'conf'} = 'show run';
$cmd{'IOS-fv'}{'page'} = 'terminal pager 0';
$cmd{'IOS-fv'}{'paus'} = '0.3';
$cmd{'CatOS'}{'prom'} = '/.+[#>]\s?\(enable\)\s*$/';
$cmd{'CatOS'}{'conf'} = 'show conf';
$cmd{'CatOS'}{'page'} = 'set length 0';
$cmd{'CatOS'}{'paus'} = '0.7';
$cmd{'Cat1900'}{'prom'} = $cmd{'IOS'}{'prom'};
$cmd{'Cat1900'}{'conf'} = 'show run';
$cmd{'Cat1900'}{'page'} = '';									# Can't be set on this oldie...
$cmd{'Cat1900'}{'paus'} = '0.3';
$cmd{'Ironware'}{'prom'} = $cmd{'IOS'}{'prom'};
$cmd{'Ironware'}{'conf'} = 'show run';
$cmd{'Ironware'}{'page'} = 'skip-page-display';
$cmd{'Ironware'}{'paus'} = '0.3';
$cmd{'ProCurve'}{'prom'} = '/.+\s?[#>]\s*$/';
$cmd{'ProCurve'}{'conf'} = 'show run';
$cmd{'ProCurve'}{'page'} = 'no page';
$cmd{'ProCurve'}{'paus'} = '0.3';
$cmd{'Nortel'}{'prom'} = '/.+\s?[#>]\s*$/';
$cmd{'Nortel'}{'conf'} = 'show run';
$cmd{'Nortel'}{'page'} = 'terminal length 0';
$cmd{'Nortel'}{'paus'} = '0.3';
$cmd{'SROS'}{'prom'} = $cmd{'IOS'}{'prom'};
$cmd{'SROS'}{'conf'} = 'show run';
$cmd{'SROS'}{'page'} = 'terminal length 0';
$cmd{'SROS'}{'macd'} = 'sh bridge | exclude \*\*\*';
$cmd{'SROS'}{'paus'} = '0.3';

#============================================================================
# Check whether SSH is available
#============================================================================
eval 'use Net::SSH::Perl';
if ($@){
	$sshnetok = 0;
	print "Net::SSH::Perl not available\n" if $main::opt{d};
}else{
	$sshnetok = 1;
	print "Net::SSH::Perl loaded\n" if $main::opt{d};
}

#============================================================================
# Bring device to enabled status
# Parameters:	current session, username, pause
# Global:		-
# Return:		0 on success, 1 on failure
#============================================================================
sub EnableDev{

	my $next  = "";										# Next expected input
	my $pre = "";
	my $block = "";

	# Alternatively this line can be used instead the select and while loop. It's probably better timing wise, but
	# causes ugly warnings in Cisco.pm I couldn't get rid of! -> Dropped Cisco 1900 support, so I could get rid of net::telnet::cisco!
	($pre, $block) = $_[0]->waitfor('/any key|Ctrl-Y|User Interface Menu|password|$misc::userlogin|[#>]\s{0,1}$/i');
	if($block =~ /any key|Ctrl-Y/i){
		print "-kick" if $main::opt{d};
		$_[0]->put("\cY");								# Since Nortel wants Ctrl-Y...
		($pre, $block) = $_[0]->waitfor('/$misc::userlogin|password|[#>]\s{0,1}$/i');
		if($block =~ /$misc::userlogin/i){
			$next = "us";
		}elsif($block =~ /password/i){
			$next = "pw";
		}
	}elsif($block =~ /User Interface Menu/){
		print "-k" if $main::opt{d};
		$_[0]->put("k");
	}elsif($block =~ /$misc::userlogin/i){
		$next = "us";
	}elsif($block =~ /password/i){
		$next = "pw";
	}
	if($next eq "us"){
		print "-USR" if $main::opt{d};
		$_[0]->print($_[1]);
		select(undef, undef, undef, $_[2]); 
		$block = $_[0]->get();
		$next = "pw";
	}
	if($next eq "pw"){
		print "-PW" if $main::opt{d};
		$_[0]->print($misc::login{$_[1]}{pw});
		select(undef, undef, undef, $_[2]); 
		$block = $_[0]->get();
	}
	if($block =~ /\n(enter )?password|$misc::userlogin|invalid|incorrect|fail/i){
		print "Tp";
		return 1;
	}else{
		if ($misc::login{$_[1]}{en} and $block =~ />\s{0,1}$/){
			print "-enable" if $main::opt{d};
			$_[0]->print('enable');
			($pre, $block) = $_[0]->waitfor('/$misc::userlogin|password|[#>]\s{0,1}$/i');
			if($block =~ /$misc::userlogin/i){
				$_[0]->print($_[1]);
				$_[0]->print($misc::login{$_[1]}{en});
			}elsif($block =~ /password/i){
				$_[0]->print($misc::login{$_[1]}{en});
			}
			($pre, $block) = $_[0]->waitfor('/#\s{0,1}$|\(enable\)\s{0,1}$/');
		}
		if ($block =~ /#\s{0,1}$|\(enable\)\s{0,1}$/){			# Make sure we are enabled now
			print "-OK" if $main::opt{d};
			return 0;			
		}else{
			print "Te";
			return 1;
		}
	}
}

#============================================================================
# Find login, if device is compatible for mac-address-table or config retrieval
# Parameters:	device name, preparation mode (fwd table or config backup)
# Global:		main::dev
# Return:		0 on success, 1 on failure, 2 on unsupported device
#============================================================================
sub PrepDev{

	my $us    = "";
	my $sshgo = 1;										# Avoid ssh...
	my $nok   = 1;
	my @users = @misc::users;

	if($_[1] eq "fwd" and !$cmd{$main::dev{$_[0]}{os}}{macd}){				# Bridge forwarding only for supported OS
		return 1;
	}
	if(defined $main::dev{$_[0]}{cp} and $main::dev{$_[0]}{cp} != 0){			# Undef -> it's new, 0 -> set to be prepd
		if(!$main::dev{$_[0]}{us}){							# Do we have a  user?
			print "Pu";
			return 2;								# No user but a real port -> failed before
		}elsif(exists $misc::login{$main::dev{$_[0]}{us}}){				# OK if in nedi.conf
			return 0;								# Lets use that then (clibad=false)
		}else{
			print "Pc";								# User not in nedi.conf -> Prep
		}
	}
	if(defined $misc::map{$main::dev{$_[0]}{ip}}{cp}){					# set telnet port
		$main::dev{$_[0]}{cp} = $misc::map{$main::dev{$_[0]}{ip}}{cp};
		print "M$main::dev{$_[0]}{cp}" if $main::opt{d};
		$sshgo = 0;									# port mapped -> avoid ssh...
	}else{
		$main::dev{$_[0]}{cp} = 23;
	}
	if($cmd{$main::dev{$_[0]}{os}}{conf}){							# won't consider potential 'fwd only' devs for now!
		do {
			$us = shift (@users);
			print " P:$us" if $main::opt{d};
			if($sshnetok and $sshgo){
				eval {
					my $ssh = Net::SSH::Perl->new($main::dev{$_[0]}{ip}, options => ["BatchMode yes",
													"RhostsAuthentication no",
													"protocol => 2" ]);
					$ssh->login($us, $misc::login{$us}{pw});
					my $sshen = ($misc::login{$us}{en})?"enable\n$misc::login{$us}{en}\n":"";
					my ($stdout, $stderr, $exit) = $ssh->cmd("","${sshen}\nexit\n");
					if ($stderr) {
						print "Ho";
						print " $stderr" if $main::opt{d};
					}else{
						if ($stdout =~ /#\s?\n|enable/){		# Make sure we were enabled
							$nok = 0;
							$main::dev{$_[0]}{cp} = 22;
							print "-OK" if $main::opt{d};
						}else{
							print "He";
							print " $stdout" if $main::opt{d};
						}
					}
				};
			}else{
				$@ = "SSH:OK=$sshnetok, GO=$sshgo";
				print "Hs";
				print " $@" if $main::opt{d};
			}
			if ($@){
				$sshgo = 0 if $@ =~ /Connection refused/;			# Don't retry SSH to save time
				my $session = Net::Telnet->new(	Host	=> $main::dev{$_[0]}{ip},
									Port	=> $main::dev{$_[0]}{cp},
									Timeout	=> $misc::timeout,
									Prompt  => $cmd{$main::dev{$_[0]}{os}}{'prom'},
									#Input_log  => "input.log",
									#Output_log  => "output.log",
									Errmode	=> 'return');
				if(defined $session){
					$nok = 0 if !&EnableDev($session, $us , $cmd{$main::dev{$_[0]}{os}}{'paus'});
					$session->close;
					select(undef, undef, undef, 0.1); 			# Pause before reconnecting to prevent hangs
				}else{
					print "Tc";
					return 2;
				}
			}
		} while ($#users ne "-1" and $nok);						# Stop if ok or we ran out of logins
	}else{
		$main::dev{$_[0]}{cp} = 0;							# Port=0 for unsupported device
		return 2;
	}
	if($nok){
		print "Tu";
		return 2;
	}else{
		print ":$main::dev{$_[0]}{cp} " if $main::opt{d};
		$main::dev{$_[0]}{us} = $us;
		return 0;
	}
}

#============================================================================
# Get Ios mac address table.
# Parameters:	device name
# Global:		misc::portprop, misc::portnew
# Return:		0 on success, 1 on failure
#============================================================================
sub BridgeFwd{

	my $line = "";
	my $nspo = 0;
	my @cam  = ();

	print " F:$main::dev{$_[0]}{us}:$main::dev{$_[0]}{cp}" if $main::opt{d};
	if($main::dev{$_[0]}{cp} == 22){
		eval {
			my $ssh = Net::SSH::Perl->new($main::dev{$_[0]}{ip}, options => ["BatchMode yes",
											"RhostsAuthentication no",
											"protocol => 2" ]);
			$ssh->login($main::dev{$_[0]}{us}, $misc::login{$main::dev{$_[0]}{us}}{pw});
			my $sshen = ($misc::login{$main::dev{$_[0]}{us}}{en})?"enable\n$misc::login{$main::dev{$_[0]}{us}}{en}\n":"";
			my $sshpg = ($cmd{$main::dev{$_[0]}{os}}{page})?"$cmd{$main::dev{$_[0]}{os}}{page}\n":"";
			my $sshms = ($misc::getfwd eq 'sec' and exists $cmd{$main::dev{$_[0]}{os}}{macs})?"$cmd{$main::dev{$_[0]}{os}}{macs}\n":"";
			my ($stdout, $stderr, $exit) = $ssh->cmd("","${sshen}${sshpg}$cmd{$main::dev{$_[0]}{os}}{macd}\n${sshms}exit\n");
			if ($stderr) {
				print "Ho";
				print " $stderr" if $main::opt{d};
				return 1;
			}else{
				@cam = split("\n", $stdout);			
			}
		};
		if ($@){
			print "Ho";
			print " $@" if $main::opt{d};
			return 1;
		}
	}else{
		my $session = Net::Telnet->new(	Host	=> $main::dev{$_[0]}{ip},
							Port	=> $main::dev{$_[0]}{cp},
							Timeout	=> $misc::timeout,
							Prompt  => $cmd{$main::dev{$_[0]}{os}}{'prom'},
							#Input_log  => "input.log",
							#Output_log  => "output.log",
							Errmode	=> 'return');
		if(defined $session){
			$session->max_buffer_length(8 * 1024 * 1024);				# Increase buffer to 8Mb
			if( &EnableDev($session, $main::dev{$_[0]}{us}, $cmd{$main::dev{$_[0]}{os}}{'paus'}) ){
				$session->close;
				return 1;
			}else{
				$session->cmd($cmd{$main::dev{$_[0]}{os}}{page}) if $cmd{$main::dev{$_[0]}{os}}{page};
				print "-$cmd{$main::dev{$_[0]}{os}}{macd}" if $main::opt{d};
				@cam = $session->cmd($cmd{$main::dev{$_[0]}{os}}{macd});
				if ($misc::getfwd eq 'sec' and exists $cmd{$main::dev{$_[0]}{os}}{macs}){
					print "-$cmd{$main::dev{$_[0]}{os}}{macs}" if $main::opt{d};
					push @cam, $session->cmd($cmd{$main::dev{$_[0]}{os}}{macs});
				}
				$session->close;
				select(undef, undef, undef, 0.1) if($main::opt{b});		# Pause if we back it up to avoid hang
			}
		}else{
			print "Tc";
			return 1;
		}
	}
	foreach my $l (@cam){
		if ($l =~ /\s+(dynamic|forward|secure(dynamic|sticky))\s+/i){			# (secure) Thanks to Duane Walker 7/2007
			my $mc = "";
			my $po = "";
			my $vl = "";
			my @mactab = split (/\s+/,$l);
			foreach my $col (@mactab){
				if ($col =~ /^(Gi|Fa|Do|Po|Vi)/){$po = &misc::Shif($col)}
				elsif ($col =~ /^[0-9|a-f]{4}\./){$mc = $col}			
				elsif ($main::dev{$_[0]}{os} ne "IOS-wl" and $col =~ /^[0-9]{1,4}$/ and !$vl){$vl = $col} # Only use this, if it's not a Cisco AP
			}
			if( exists($misc::portprop{$_[0]}{$po}) ){				# Make sure IFidx exists before using IF
				if($po =~ /\.[0-9]/){						# Does it look like a subinterface?
					my @subpo = split(/\./,$po);
					$vl = $subpo[1];
					if($misc::portprop{$_[0]}{$subpo[0]}{upl}){$misc::portprop{$_[0]}{$po}{upl} = 1}	# inherit uplink metric on subinterface
				}
				$mc =~ s/\.//g;

				if ($vl !~ /$misc::ignoredvlans/){
					if ($po =~ /^.EC-|^Po[0-9]|channel/){
						$misc::portprop{$_[0]}{$po}{chn} = 1;
					}
					$misc::portprop{$_[0]}{$po}{pop}++;
					$misc::portnew{$mc}{$_[0]}{po} = $po;
					$misc::portnew{$mc}{$_[0]}{vl} = $vl;
					print "\n FWC:$mc on $po vl$vl" if $main::opt{v};
					$nspo++;
				}
			}
		}
	}
	print "f$nspo";
	return 0;
}

#============================================================================
# Fetch  Config.
# Parameters:	device name
# Global:		misc::curcfg
# Return:		0 on success, error on failure
#============================================================================
sub Config{

	my $en        = "";
	my $go        = 0;
	my @run       = ();
	
	print "B:$main::dev{$_[0]}{us}:$main::dev{$_[0]}{cp}" if $main::opt{d};
	if($main::dev{$_[0]}{cp} == 22){
		eval {
			my $ssh = Net::SSH::Perl->new($main::dev{$_[0]}{ip}, options => ["BatchMode yes",
											"RhostsAuthentication no",
											"protocol => 2" ]);
			$ssh->login($main::dev{$_[0]}{us}, $misc::login{$main::dev{$_[0]}{us}}{pw});
			my $sshen = ($misc::login{$main::dev{$_[0]}{us}}{en})?"enable\n$misc::login{$main::dev{$_[0]}{us}}{en}\n":"";
			my $sshpg = ($cmd{$main::dev{$_[0]}{os}}{page})?"$cmd{$main::dev{$_[0]}{os}}{page}\n":"";
			print "-$cmd{$main::dev{$_[0]}{os}}{page}-$cmd{$main::dev{$_[0]}{os}}{conf}" if $main::opt{d};
			my ($stdout, $stderr, $exit) = $ssh->cmd("","${sshen}${sshpg}$cmd{$main::dev{$_[0]}{os}}{conf}\nexit\n");
			if ($stderr) {
				print "Ho";
				print " $stderr" if $main::opt{d};
				return "SSH:$stderr";
			}else{
				@run = split("\n", $stdout);			
			}
		};
		if ($@){
			print "Hs";
			print " $@" if $main::opt{d};
			return "SSH:$@";
		}
	}else{
		my $session = Net::Telnet->new(	Host	=> $main::dev{$_[0]}{ip},
							Port	=> $main::dev{$_[0]}{cp},
							Timeout	=> $misc::timeout,
							Prompt  => $cmd{$main::dev{$_[0]}{os}}{'prom'},
							#Input_log  => "input.log",
							#Output_log  => "output.log",
							Errmode	=> 'return');
		if(defined $session){
			$session->max_buffer_length(8 * 1024 * 1024);				# Increase buffer to 8Mb
			if( &EnableDev($session, $main::dev{$_[0]}{us}, $cmd{$main::dev{$_[0]}{os}}{'paus'}) ){
				$session->close;
				return "Coudn't enable!";
			}else{
				$session->timeout($misc::timeout * 10);				# Increase for building config
				print "-$cmd{$main::dev{$_[0]}{os}}{page}-$cmd{$main::dev{$_[0]}{os}}{conf}" if $main::opt{d};
				$session->cmd($cmd{$main::dev{$_[0]}{os}}{page}) if $cmd{$main::dev{$_[0]}{os}}{page};
				@run = $session->cmd($cmd{$main::dev{$_[0]}{os}}{conf});
				$session->close;
			}
		}else{
			print "Tc";
			return "Telnet port $main::dev{$_[0]}{cp} failed!";
		}
	}
	foreach my $line (@run){
		$line =~ s/[\r\n]//g;
		if ($line =~ /^(Running|Current|PIX|FWSM|ASA|Building)\s|^begin$|\*\*\* CORE|<config>/){$go = 1}
		if ($go and $line !~ /--More--|^$/){						# --More-- should only be seen on Cat1900s
			print "\n CFG:$line" if $main::opt{v};
			push @misc::curcfg,$line;
		}
		if ($line =~ /^(: )?end$|<\/config>/){$go = 0}
	}
	if( scalar(@misc::curcfg) < 2 ){ 
		print "Be";
		return "Empty config!";
	}else{
		print "-".scalar(@misc::curcfg) if $main::opt{d};
		return "";
	}
}

#============================================================================
# Send commands to device (used by the GUI helper Devsend.pl)
# Parameters:	IP, port, user, pass, OS, command file
# Global:		-
# Return:		-
#============================================================================
sub SendCmd{

	my $ip = $ARGV[0];
	my $po = $ARGV[1];
	my $us = $ARGV[2];
	my $pw = $ARGV[3];
	my $os = $ARGV[4];
	my $cf = $ARGV[5];

	if($misc::guiauth =~ /-pass$/){
		$misc::login{$us}{pw} = $pw;
	}

	open  (CFG, "$cf" );
	my @cmd = <CFG>;
	close(CFG);
	chomp @cmd;

	print "$ip:$po $os $us/$misc::login{$us}{pw} -> $cf(". scalar @cmd .") T:$misc::timeout s\n"  if $main::opt{d};

	open  (LOG, ">$cf-$ip.log" ) or print " can't write to $cf-$ip.log";

	my $session = Net::Telnet->new(	Host	=> $ip,
					Port	=> $po,
					Prompt  => '/.+?[#>]\s?(?:\(enable\)\s*)?$/',
					Timeout => $misc::timeout,
					Errmode => "return",
					#Input_log  => "log/input.log",
					#Output_log  => "log/output.log",
					);
	if(defined $session){
		$session->max_buffer_length(8 * 1024 * 1024);							# Increase buffer to 8Mb
		if( &EnableDev($session,$us, $cmd{$os}{'paus'}) ){
			$session->close;
		}else{
			print " (".localtime($main::now).") ";
			$session->cmd($cmd{$os}{'page'}) if $cmd{$os}{'page'};
			foreach my $c (@cmd){
				print LOG "$c\n";
				print LOG join("", $session->cmd($c) );
				print ".";
				if( $session->errmsg ){
					print LOG $session->errmsg;
					print $session->errmsg;
					last;
				}
			}
			close (LOG);
			$session->close;
		}
	}else{
		$session->close;
		close (LOG);
		print " telnet $ip:$po " . $session->errmsg;
	}
}

1;
