#============================================================================
#
# Program: libmsq.pl
# Programmer: Remo Rickli, Paul Venezia, Steffen Neuser
#
# -> Functions for the MYSQL Database <-
#
#============================================================================
package db;

use DBI;

#===================================================================
# Connect as admin, drop existing DB and create nedi db and user first
# Parameters:	$adminuser,$adminpass,$nedihost
# Global:	-
# Return:	-
#===================================================================
sub InitDB{

	$dbh = DBI->connect("DBI:mysql:mysql:$misc::dbhost", "$_[0]", "$_[1]", { RaiseError => 1, AutoCommit => 1});
	my $mysqlVer;
	my $sth = $dbh->prepare("SELECT VERSION()");
	$sth->execute();
	while ((my @f) = $sth->fetchrow) {
		$mysqlVer = $f[0];
	}
	print "MySQL Version	: $mysqlVer\n";
	print "----------------------------------------------------------------------\n";
	$dbh->do("DROP DATABASE IF EXISTS $misc::dbname");
	print "Old DB $misc::dbname dropped!\n";

	print "Creating $misc::dbname ";
	$dbh->do("CREATE DATABASE $misc::dbname");
	$dbh->do("GRANT ALL PRIVILEGES ON $misc::dbname.* TO \'$misc::dbuser\'\@\'$_[2]\' IDENTIFIED BY \'$misc::dbpass\'");
	if ($mysqlVer =~ /5\./) {                #fix for mysql 5.0 with old client libs
		$dbh->do("SET PASSWORD FOR \'$misc::dbuser\'\@\'$_[2]\' = OLD_PASSWORD(\'$misc::dbpass\')");
	}
	print "for $misc::dbuser\@$_[2]\n";
	$sth->finish if $sth;
	$dbh->disconnect();

#---Connect as nedi db user and create tables.
	$dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});

	print "Creating Tables: ";
	print "devices, ";
	$dbh->do("CREATE TABLE devices	(	name VARCHAR(64) BINARY NOT NULL UNIQUE, ip INT unsigned, serial VARCHAR(32), type VARCHAR(32),
						firstseen INT unsigned, lastseen INT unsigned, services TINYINT unsigned,
						description VARCHAR(255), os VARCHAR(16), bootimage VARCHAR(64),
						location VARCHAR(255), contact VARCHAR(255),
						vtpdomain VARCHAR(32), vtpmode TINYINT unsigned, snmpversion TINYINT unsigned,
						community VARCHAR(32), cliport SMALLINT unsigned, login VARCHAR(32),
						icon VARCHAR(16), origip INT unsigned, cpu TINYINT unsigned,memcpu INT unsigned,
						temp TINYINT unsigned, cusvalue INT unsigned, cuslabel VARCHAR(32),
						sysobjid VARCHAR(255), logalarm VARCHAR(255), flags TINYINT unsigned,
						index (name(8)), PRIMARY KEY  (name) )");
 	$dbh->commit;

	print "devdel, ";
	$dbh->do("CREATE TABLE devdel	(	device VARCHAR(64) BINARY NOT NULL UNIQUE, user VARCHAR(32), time INT unsigned,
						index (device(8)), PRIMARY KEY  (device) )");
 	$dbh->commit;

	print "modules, ";
	$dbh->do("CREATE TABLE modules	(	device VARCHAR(64) BINARY, slot VARCHAR(64), model VARCHAR(32), description VARCHAR(255),
						serial VARCHAR(32), hw VARCHAR(128), fw VARCHAR(128), sw VARCHAR(128),
						modidx SMALLINT unsigned, index (device(8)), index (slot(8)) ) ");
 	$dbh->commit;

	print "interfaces, ";
	$dbh->do("CREATE TABLE interfaces(	device VARCHAR(64) BINARY, ifname VARCHAR(32), ifidx BIGINT unsigned,
						linktype CHAR(4), type INT unsigned, mac CHAR(12),
						description VARCHAR(255), alias VARCHAR(64), status TINYINT unsigned,
						speed BIGINT unsigned, duplex CHAR(2), vlid SMALLINT unsigned,
						inoct BIGINT unsigned, inerr INT unsigned, outoct BIGINT unsigned, outerr INT unsigned,
						dinoct BIGINT signed, dinerr INT signed, doutoct BIGINT signed, douterr INT signed,
						comment VARCHAR(255), index (device(8)), index (ifname(8)),index (ifidx) )");
 	$dbh->commit;

	print "networks, ";
	$dbh->do("CREATE TABLE networks (	device VARCHAR(64) BINARY, ifname VARCHAR(32), ip INT unsigned, mask INT unsigned,
						vrfname VARCHAR(32), status TINYINT unsigned, index (device(8)), index (ifname), index (ip) )");
 	$dbh->commit;

	print "configs, ";
	$dbh->do("CREATE TABLE configs	(	device VARCHAR(64) BINARY NOT NULL UNIQUE, config MEDIUMTEXT, changes MEDIUMTEXT ,
						time INT unsigned, index (device(8)), PRIMARY KEY  (device)  )");
 	$dbh->commit;

	print "stock, ";
	$dbh->do("CREATE TABLE stock	(	serial VARCHAR(32) UNIQUE, type VARCHAR(32),user VARCHAR(32),
						time INT unsigned, location VARCHAR(255), state TINYINT unsigned, comment VARCHAR(255),
						lastseen INT unsigned, source VARCHAR(32) default '-', index(serial) )");
 	$dbh->commit;

	print "vlans, ";
	$dbh->do("CREATE TABLE vlans	(	device VARCHAR(64) BINARY, vlanid SMALLINT unsigned,
						vlanname VARCHAR(32), index(vlanid) )");
 	$dbh->commit;

	print "links, ";
	$dbh->do("CREATE TABLE links	(	id INT unsigned NOT NULL AUTO_INCREMENT, device VARCHAR(64) BINARY,
						ifname VARCHAR(32), neighbour VARCHAR(64) BINARY, nbrifname VARCHAR(32),
						bandwidth BIGINT unsigned, type CHAR(4), power INT unsigned, nbrduplex CHAR(2),
						nbrvlanid SMALLINT unsigned, index (id), index (device(8)), index (ifname(8)),
						index (neighbour(8)), index (nbrifname(8)), PRIMARY KEY  (id) )");
 	$dbh->commit;

	print "monitoring, ";
	$dbh->do("CREATE TABLE monitoring(	device VARCHAR(64) BINARY NOT NULL UNIQUE, status INT unsigned, depend VARCHAR(64),
						sms INT unsigned, mail INT unsigned, lastchk INT unsigned, uptime INT unsigned,
						lost INT unsigned, ok INT unsigned, delay SMALLINT unsigned, index (device(8)) )");
 	$dbh->commit;

	print "messages, ";
	$dbh->do("CREATE TABLE messages(	id INT unsigned NOT NULL AUTO_INCREMENT, level TINYINT unsigned, time INT unsigned,
						source VARCHAR(64), info VARCHAR(255), index (id), index (source(8)), index (level),
						index (time), PRIMARY KEY  (id) )");
 	$dbh->commit;

	print "incidents, ";
	$dbh->do("CREATE TABLE incidents(	id INT unsigned NOT NULL AUTO_INCREMENT, level TINYINT unsigned, device VARCHAR(64),
						deps INT unsigned, firstseen INT unsigned, lastseen INT unsigned, who VARCHAR(32),
						time INT unsigned, category TINYINT unsigned, comment VARCHAR(255),
						index (id), PRIMARY KEY  (id) )");
 	$dbh->commit;

	print "locations, ";
	$dbh->do("CREATE TABLE locations(	id INT unsigned NOT NULL AUTO_INCREMENT,region VARCHAR(32) BINARY NOT NULL,
						city VARCHAR(32), building VARCHAR(32), x SMALLINT unsigned, y SMALLINT unsigned,
						comment VARCHAR(64), index(region),PRIMARY KEY  (id)  )");
 	$dbh->commit;

	print "nodes, ";
	$dbh->do("CREATE TABLE nodes 	(	name VARCHAR(64), ip INT unsigned, mac CHAR(12) NOT NULL UNIQUE, oui VARCHAR(32),
						firstseen INT unsigned, lastseen INT unsigned, device VARCHAR(64) BINARY,
						ifname VARCHAR(32), vlanid SMALLINT unsigned, ifmetric TINYINT unsigned,
						ifupdate INT unsigned, ifchanges INT unsigned,	ipupdate INT unsigned,
						ipchanges INT unsigned, iplost INT unsigned, arp SMALLINT unsigned,
						tcpports VARCHAR(64), udpports VARCHAR(64), os VARCHAR(32) default '-',
						type VARCHAR(32) default '-', osupdate INT unsigned default 0,
						index (name(8)), index(ip), index(mac), index(vlanid), PRIMARY KEY  (mac) )");
 	$dbh->commit;
	print "iftrack, ";
	$dbh->do("CREATE TABLE iftrack	(	mac CHAR(12),ifupdate INT unsigned,
						device VARCHAR(64) BINARY, ifname VARCHAR(32), vlanid SMALLINT unsigned,
						ifmetric TINYINT unsigned, index(mac), index(ifupdate) )");
 	$dbh->commit;
	print "iptrack, ";
	$dbh->do("CREATE TABLE iptrack (	mac CHAR(12),ipupdate INT unsigned,
						name VARCHAR(64),ip INT unsigned,index(mac),index(ipupdate) )");
 	$dbh->commit;

	print "stolen, ";
	$dbh->do("CREATE TABLE stolen 	(	name VARCHAR(64), ip INT unsigned, mac CHAR(12) UNIQUE, device VARCHAR(64) BINARY,
						ifname VARCHAR(32), who VARCHAR(32), time INT unsigned, index(mac), PRIMARY KEY  (mac) )");
 	$dbh->commit;

	print "user, ";
	$dbh->do("CREATE TABLE user 	(	name VARCHAR(32) NOT NULL UNIQUE, password VARCHAR(32) NOT NULL default '',
						groups TINYINT unsigned NOT NULL default '0', email VARCHAR(64), phone VARCHAR(32),
						time INT unsigned, lastseen INT unsigned, comment VARCHAR(255),
						language VARCHAR(16) NOT NULL default 'english', theme VARCHAR(16) NOT NULL default 'default',
						volume TINYINT unsigned NOT NULL default '10', columns TINYINT unsigned NOT NULL default '5',
						msglimit TINYINT unsigned NOT NULL default '5', graphs TINYINT unsigned NOT NULL default '2',
						PRIMARY KEY (name) )");
	$sth = $dbh->prepare("INSERT INTO user (name,password,groups,time,comment,volume,columns,msglimit,graphs) VALUES ( ?,?,?,?,?,?,?,?,? )");
	$sth->execute ( 'admin','21232f297a57a5a743894a0e4a801fc3','255',$main::now,'default admin','100','8','10','3' );
 	$dbh->commit;

	print "system, ";
	$dbh->do("CREATE TABLE system	(	name VARCHAR(32) NOT NULL UNIQUE, value VARCHAR(32) default '',
						index(name) )");
	$sth = $dbh->prepare("INSERT INTO system (name,value) VALUES ( ?,? )");
	$sth->execute ( 'nodlock','0' );
	$sth->execute ( 'threads','0' );
 	$dbh->commit;

	print "chat, ";
	$dbh->do("CREATE TABLE chat (time INT unsigned,  user VARCHAR(32), message VARCHAR(255), index(time),index (user(8)) )");

	print "wlan";
	$dbh->do("CREATE TABLE wlan (mac VARCHAR(12),time INT unsigned, index(mac) )");
	my @wlan = ();
	if (-e "$main::p/inc/wlan.txt"){
		open  ("WLAN", "$main::p/inc/wlan.txt" );
		@wlan = <WLAN>;
		close("WLAN");
		chomp(@wlan);
	}
	$sth = $dbh->prepare("INSERT INTO wlan (mac,time) VALUES ( ?,? )");
	for my $mc (sort @wlan ){ $sth->execute ( $mc,$main::now ) }
 	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect();
	print " done.\n";
}

#===================================================================
# Read Previous device table.
# Parameters:	(optional column, values to match)
# Global:	main::dev
# Return:	-
#===================================================================
sub ReadDev {

	my $npdev = 0;
	my $where = "";

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	if($_[0] and $_[1]){$where = "WHERE $_[0] = \"$_[1]\""}
	my $sth = $dbh->prepare("SELECT * FROM devices $where");
	$sth->execute();
	if($sth->rows){
		undef (%main::dev);								# Replace entries only when we got some new ones. Avoid problems in moni.pl while nedi.pl updates devices
	}
	while ((my @f) = $sth->fetchrow_array) {
		$main::dev{$f[0]}{ip} = &misc::Dec2Ip($f[1]);
		$main::dev{$f[0]}{oi} = &misc::Dec2Ip($f[19]);					# Used for community tracking too
		$main::dev{$f[0]}{sn} = $f[2];
		$main::dev{$f[0]}{ty} = $f[3];
		$main::dev{$f[0]}{fs} = $f[4];
		$main::dev{$f[0]}{ls} = $f[5];
		$main::dev{$f[0]}{sv} = $f[6];
		$main::dev{$f[0]}{de} = $f[7];
		$main::dev{$f[0]}{os} = $f[8];
		$main::dev{$f[0]}{bi} = $f[9];
		$main::dev{$f[0]}{lo} = $f[10];
		$main::dev{$f[0]}{co} = $f[11];
		$main::dev{$f[0]}{vd} = $f[12];
		$main::dev{$f[0]}{vm} = $f[13];
		$main::dev{$f[0]}{sp} = $f[14] & 127;
		$main::dev{$f[0]}{hc} = $f[14] & 128;
		$main::dev{$f[0]}{cm} = $f[15];
		$main::dev{$f[0]}{cp} = $f[16];
		$main::dev{$f[0]}{us} = $f[17];
		$main::dev{$f[0]}{ic} = $f[18];
		$main::dev{$f[0]}{so} = $f[25];
		$main::dev{$f[0]}{mi} = $f[26];
		$main::dev{$f[0]}{fl} = $f[27];
		$misc::snmpini{$main::dev{$f[0]}{ip}}{sp} = $main::dev{$f[0]}{sp};		# Tie comm & ver to IP,
		$misc::snmpini{$main::dev{$f[0]}{ip}}{cm} = $main::dev{$f[0]}{cm};
		$misc::snmpini{$main::dev{$f[0]}{oi}}{sp} = $main::dev{$f[0]}{sp};		# it's all we have at first
		$misc::snmpini{$main::dev{$f[0]}{oi}}{cm} = $main::dev{$f[0]}{cm};
		$npdev++;
	}
	$sth->finish if $sth;
	$dbh->disconnect;
	print "DEV:	$npdev devices read from $misc::dbname.devices\n" if (!$main::opt{D});	# Don't print if daemonized in sylog and moni
}

#===================================================================
# Read Links
# Parameters:	(optional column, values to match)
# Global:	main::link
# Return:	-
#===================================================================
sub ReadLink {

	my $nlink = 0;
	my $where = "";

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	if($_[0] and $_[1]){$where = "WHERE $_[0] = \"$_[1]\""}
	my $sth = $dbh->prepare("SELECT * FROM links $where");
	$sth->execute();
	while ((my @l) = $sth->fetchrow_array) {
		$main::link{$l[1]}{$l[2]}{$l[3]}{$l[4]}{bw} = $l[5];
		$main::link{$l[1]}{$l[2]}{$l[3]}{$l[4]}{ty} = $l[6];
		$main::link{$l[1]}{$l[2]}{$l[3]}{$l[4]}{pr} = $l[7];
		$main::link{$l[1]}{$l[2]}{$l[3]}{$l[4]}{du} = $l[8];
		$main::link{$l[1]}{$l[2]}{$l[3]}{$l[4]}{vl} = $l[9];
		$nlink++;
	}
	$sth->finish if $sth;
	$dbh->disconnect;
	print "LINK:	$nlink links ($where) read from $misc::dbname.links\n";
}

#===================================================================
# Read Previous node table.
# Parameters:	-
# Global:	main::nod
# Return:	-
#===================================================================
sub ReadNod {

	my $nnod = 0;
	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $sth = $dbh->prepare("SELECT * FROM nodes $_[0]");
	$sth->execute();
	while ((my @f) = $sth->fetchrow_array) {
		$main::nod{$f[2]}{na} = $f[0];
		$main::nod{$f[2]}{ip} = &misc::Dec2Ip($f[1]);
		$main::nod{$f[2]}{nv} = $f[3];
		$main::nod{$f[2]}{fs} = $f[4];
		$main::nod{$f[2]}{ls} = $f[5];
		$main::nod{$f[2]}{dv} = $f[6];
		$main::nod{$f[2]}{if} = $f[7];
		$main::nod{$f[2]}{vl} = $f[8];
		$main::nod{$f[2]}{im} = $f[9];
		$main::nod{$f[2]}{iu} = $f[10];
		$main::nod{$f[2]}{ic} = $f[11];
		$main::nod{$f[2]}{au} = $f[12];
		$main::nod{$f[2]}{ac} = $f[13];
		$main::nod{$f[2]}{al} = $f[14];
		$main::nod{$f[2]}{an} = $f[15];
		$main::nod{$f[2]}{tp} = $f[16];
		$main::nod{$f[2]}{up} = $f[17];
		$main::nod{$f[2]}{os} = $f[18];
		$main::nod{$f[2]}{ty} = $f[19];
		$main::nod{$f[2]}{ou} = $f[20];
		$nnod++;
	}
	$sth->finish if $sth;
	$dbh->disconnect;
	print "Node:	$nnod nodes read ($_[0]) from $misc::dbname.nodes\n";
}

#===================================================================
# Backup configuration and any changes.
# Parameters:	device name
# Global:	-
# Return:	-
#===================================================================
sub BackupCfg {

	my $chg  = "";
	my $dv = $_[0];

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $sth = $dbh->prepare("SELECT config,changes FROM configs where device = \"$dv\"");
	$sth->execute();

	if($sth->rows == 0 and !$main::opt{t}){										# No previous config found, therefore write new.
		$sth = $dbh->prepare("INSERT INTO configs(device,config,changes,time) VALUES ( ?,?,?,? )");
		$sth->execute ($dv,join("\n",@misc::curcfg),$chg,$main::now);
		&misc::WriteCfg($dv) if $main::opt{B};
		print "Bn";
	}elsif($sth->rows == 1){											# Previous config found, get changes
		my @pc = $sth->fetchrow_array;
		my @pcfg = split(/\n/,$pc[0]);
		my $achg = &misc::GetChanges(\@pcfg, \@misc::curcfg);
		if(!$main::opt{t}){
			if($achg){											# Only write new, if changed
				$chg  = $pc[1] . "#--- " . localtime($main::now) ." ---#\n". $achg;
				$dbh->do("DELETE FROM configs where device = \"$dv\"");
				$sth = $dbh->prepare("INSERT INTO configs(device,config,changes,time) VALUES ( ?,?,?,? )");
				$sth->execute ($dv,join("\n",@misc::curcfg),$chg,$main::now);
				print "Bu";
				&misc::WriteCfg($dv) if $main::opt{B};
				if($misc::notify =~ /c/){
					my $len = length($achg);
					$sth = $dbh->prepare("INSERT INTO messages(level,time,source,info) VALUES ( ?,?,?,? )");
					$sth->execute ('50',$main::now,$dv,"Config changed ($len chars)");
				}
			} else {											# Write config file in case it was lost before
			    &misc::WriteCfg($dv) if $main::opt{B} and ! -e "$misc::nedipath/conf/$dv";
			}
		}
	}
	$sth->finish if $sth;
	$dbh->disconnect;
}

#===================================================================
# Write a device
# Parameters:	devicename
# Global:	-
# Return:	-
#===================================================================
sub WriteDev {

	my $na = $_[0];

	if(!defined $main::dev{$na}{cm}){$main::dev{$na}{cm}	= ""}
	if(!defined $main::dev{$na}{us}){$main::dev{$na}{us}	= ""}
	if(!defined $main::dev{$na}{cp}){$main::dev{$na}{cp}	= 0}
	if(!defined $main::dev{$na}{sp}){$main::dev{$na}{sp}	= 0}
	if(!defined $main::dev{$na}{hc}){$main::dev{$na}{hc}	= 0}
	if(!defined $main::dev{$na}{oi}){$main::dev{$na}{oi}	= 0}
	if(!$main::dev{$na}{ic}){
		if($main::dev{$na}{sv} > 8){
			$main::dev{$na}{ic} = 'csan';
		}elsif($main::dev{$na}{sv} > 4){
			$main::dev{$na}{ic} = 'w3an';
		}elsif($main::dev{$na}{sv} > 1){
			$main::dev{$na}{ic} = 'w2an';
		}else{
			$main::dev{$na}{ic} = 'w1an';
		}
	}
	my $sphc = $main::dev{$na}{sp} + $main::dev{$na}{hc};
	my $dip = &misc::Ip2Dec($main::dev{$na}{ip});
	my $doi = &misc::Ip2Dec($main::dev{$na}{oi});

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	$dbh->do("DELETE FROM  devices where name = \"$na\"");
	$sth = $dbh->prepare("INSERT INTO devices(	name,ip,serial,type,firstseen,lastseen,services,
							description,os,bootimage,location,contact,
							vtpdomain,vtpmode,snmpversion,community,cliport,login,
							icon,origip,cpu,memcpu, temp, cusvalue, cuslabel, sysobjid, logalarm, flags
							) VALUES ( ?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,? )");
	$sth->execute (	$na,
			$dip,
			$main::dev{$na}{sn},
			$main::dev{$na}{ty},
			$main::dev{$na}{fs},
			$main::dev{$na}{ls},
			$main::dev{$na}{sv},
			$main::dev{$na}{de},
			$main::dev{$na}{os},
			$main::dev{$na}{bi},
			$main::dev{$na}{lo},
			$main::dev{$na}{co},
			$main::dev{$na}{vd},
			$main::dev{$na}{vm},
			$sphc,
			$main::dev{$na}{cm},
			$main::dev{$na}{cp},
			$main::dev{$na}{us},
			$main::dev{$na}{ic},
			$doi,
			$main::dev{$na}{cpu},
			$main::dev{$na}{mcp},
			$main::dev{$na}{tmp},
			$main::dev{$na}{cuv},
			$main::dev{$na}{cul},
			$main::dev{$na}{so},
			$main::dev{$na}{mi},
			$main::dev{$na}{fl}
			);
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
}

#===================================================================
# Write the interfaces table
# Parameters:	devicename
# Global:	main::int
# Return:	-
#===================================================================
sub WriteInt {

	my $nint = 0;
	my $nwar = 0;
	my $nalr = 0;

	my $dv = $_[0];

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	my $sth = $dbh->prepare("SELECT * FROM interfaces WHERE device = \"$dv\"");
	$sth->execute();
	while ((my @f) = $sth->fetchrow_array) {
		if( exists $main::int{$f[0]}{$f[2]} ){
			$main::int{$f[0]}{$f[2]}{dio} = $main::int{$f[0]}{$f[2]}{ioc} - $f[12];
			if ($main::int{$f[0]}{$f[2]}{dio} < 0){
				$main::int{$f[0]}{$f[2]}{dio} = 0;
				$main::int{$f[0]}{$f[2]}{ioc} = 71;
			}
			$main::int{$f[0]}{$f[2]}{die} = $main::int{$f[0]}{$f[2]}{ier} - $f[13];
			if ($main::int{$f[0]}{$f[2]}{die} < 0){
				$main::int{$f[0]}{$f[2]}{die} = 0;
			}
			$main::int{$f[0]}{$f[2]}{doo} = $main::int{$f[0]}{$f[2]}{ooc} - $f[14];
			if ($main::int{$f[0]}{$f[2]}{doo} < 0){
				$main::int{$f[0]}{$f[2]}{doo} = 0;
				$main::int{$f[0]}{$f[2]}{ooc} = 71;
			}
			$main::int{$f[0]}{$f[2]}{doe} = $main::int{$f[0]}{$f[2]}{oer} - $f[15];
			if ($main::int{$f[0]}{$f[2]}{doe} < 0){
				$main::int{$f[0]}{$f[2]}{doe} = 0;
			}
			if($main::int{$f[0]}{$f[2]}{sta} != 1 and $f[8] == 1){
				$main::int{$f[0]}{$f[2]}{sta} = $main::int{$f[0]}{$f[2]}{sta} + 128;
				$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"$main::int{$f[0]}{$f[2]}{ina} Disabled!\")") if($misc::notify =~ /t/);
			}
			if($main::int{$f[0]}{$f[2]}{lty} ne $f[3]){
				$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"Linkchange ".(($f[3])?"from $f[3] ":"").(($main::int{$f[0]}{$f[2]}{lty})?"to $main::int{$f[0]}{$f[2]}{lty} ":"")."detected on $main::int{$f[0]}{$f[2]}{ina}!\")") if($misc::notify =~ /d/);
			}
		}
	}
	$sth->finish if $sth;
	$dbh->do("DELETE FROM  interfaces where device = \"$dv\"");
	$sth = $dbh->prepare("INSERT INTO interfaces(	device,ifname,ifidx,linktype,type,mac,description,alias,status,speed,duplex,vlid,
							inoct,inerr,outoct,outerr,dinoct,dinerr,doutoct,douterr,comment)
							VALUES ( ?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,? )");
	foreach my $i ( sort keys %{$main::int{$dv}} ){
		$sth->execute (	$dv,
				$main::int{$dv}{$i}{ina},
				$i,
				$main::int{$dv}{$i}{lty},
				$main::int{$dv}{$i}{typ},
				$main::int{$dv}{$i}{mac},
				$main::int{$dv}{$i}{des},
				$main::int{$dv}{$i}{ali},
				$main::int{$dv}{$i}{sta},
				$main::int{$dv}{$i}{spd},
				$main::int{$dv}{$i}{dpx},
				$main::int{$dv}{$i}{vln},
				$main::int{$dv}{$i}{ioc},
				$main::int{$dv}{$i}{ier},
				$main::int{$dv}{$i}{ooc},
				$main::int{$dv}{$i}{oer},
				$main::int{$dv}{$i}{dio},
				$main::int{$dv}{$i}{die},
				$main::int{$dv}{$i}{doo},
				$main::int{$dv}{$i}{doe},
				$main::int{$dv}{$i}{com} );
		$nint++;
		if($misc::notify =~ /t/ and $main::int{$dv}{$i}{spd}){
			my $rioct = int( $main::int{$dv}{$i}{dio} * 800 / ($misc::rrdstep * $main::int{$dv}{$i}{spd}) );
			my $rooct = int( $main::int{$dv}{$i}{doo} * 800 / ($misc::rrdstep * $main::int{$dv}{$i}{spd}) );
			if($rioct > $misc::trfa and $rioct < 101){
				$dbh->do("INSERT INTO messages (level,time,source,info) values (200,$main::now,\"$dv\",\"Average inbound traffic on $main::int{$dv}{$i}{ina} is $rioct% for ${misc::rrdstep}s!\")");
				$nalr++;
			}elsif($rioct > $misc::trfw and $rioct < 101){
				$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"Average inbound traffic on $main::int{$dv}{$i}{ina} is $rioct% for ${misc::rrdstep}s\")");
				$nwar++;
			}
			if($rooct > $misc::trfa and $rooct < 101){
				$dbh->do("INSERT INTO messages (level,time,source,info) values (200,$main::now,\"$dv\",\"Average outbound traffic on $main::int{$dv}{$i}{ina} is $rioct% for ${misc::rrdstep}s!\")");
				$nalr++;
			}elsif($rooct > $misc::trfw and $rooct < 101){
				$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"Average outbound traffic on $main::int{$dv}{$i}{ina} is $rioct% for ${misc::rrdstep}s\")");
				$nwar++;
			}
			if($main::int{$dv}{$i}{typ} != 71){								# Ignore Wlan IF
				if($main::int{$dv}{$i}{die} > $misc::rrdstep){
					$dbh->do("INSERT INTO messages (level,time,source,info) values (200,$main::now,\"$dv\",\"$main::int{$dv}{$i}{die} inbound errors on $main::int{$dv}{$i}{ina} in ${misc::rrdstep}s!\")");
					$nalr++;
				}elsif($main::int{$dv}{$i}{die} > $misc::rrdstep / 60){
					$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"$main::int{$dv}{$i}{die} inbound errors on $main::int{$dv}{$i}{ina} in ${misc::rrdstep}s\")");
					$nwar++;
				}
				if($main::int{$dv}{$i}{doe} > $misc::rrdstep){
					$dbh->do("INSERT INTO messages (level,time,source,info) values (200,$main::now,\"$dv\",\"$main::int{$dv}{$i}{doe} outbound errors on $main::int{$dv}{$i}{ina} in ${misc::rrdstep}s!\")");
					$nalr++;
				}elsif($main::int{$dv}{$i}{doe} > $misc::rrdstep / 60){
					$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"$main::int{$dv}{$i}{doe} outbound errors on $main::int{$dv}{$i}{ina} in ${misc::rrdstep}s\")");
					$nwar++;
				}
			}
		}
	}
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
	print "IF:	$nint interfaces written to $misc::dbname.interfaces\n" if $main::opt{v};
	print "IF:	$nwar warnings/$nalr alerts written to $misc::dbname.messages\n" if $main::opt{v} and ($nwar or $nalr);
}

#===================================================================
# Write the modules table
# Parameters:	devicename
# Global:	-
# Return:	-
#===================================================================
sub WriteMod {

	my $nmod = 0;
	my %dbmod = ();

	my $dv = $_[0];

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	if($misc::notify =~ /m/){								# Track existing mods if enabled
		my $sth = $dbh->prepare("SELECT * FROM modules WHERE device = \"$dv\"");
		$sth->execute();
		while ((my @f) = $sth->fetchrow_array) {
			if(exists $main::mod{$dv}{$f[8]}){					# Check idx to avoid defining entry..
				if($f[3] != $main::mod{$dv}{$f[8]}{de}){			# ..this would define it accidentaly
					$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"Module $f[3] SN:$f[4] Slot:$f[1] replaced by different model\")");
				}elsif($f[4] != $main::mod{$dv}{$f[8]}{sn}){
					$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"Module $f[3] SN:$f[4] Slot:$f[1] replaced by same model\")");
				}
			}else{
				$dbh->do("INSERT INTO messages (level,time,source,info) values (150,$main::now,\"$dv\",\"Module $f[3] SN:$f[4] Slot:$f[1] removed\")");
			}
		}

	}
	$sth->finish if $sth;
	$dbh->do("DELETE FROM  modules where device = \"$dv\"");
	my $sth = $dbh->prepare("INSERT INTO modules(device,slot,model,description,serial,hw,fw,sw,modidx) VALUES ( ?,?,?,?,?,?,?,?,? )");
	foreach my $i ( sort keys %{$main::mod{$dv}} ){
		$sth->execute (	$dv,
				$main::mod{$dv}{$i}{sl},
				$main::mod{$dv}{$i}{mo},
				$main::mod{$dv}{$i}{de},
				$main::mod{$dv}{$i}{sn},
				$main::mod{$dv}{$i}{hw},
				$main::mod{$dv}{$i}{fw},
				$main::mod{$dv}{$i}{sw},
				$i );
		$nmod++;
	}
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
	print "MOD:	$nmod modules written to $misc::dbname.modules\n" if $main::opt{v};
}

#===================================================================
# Write the vlan table
# Parameters:	devicename
# Global:	-
# Return:	-
#===================================================================
sub WriteVlan {

	my $nvlans = 0;

	my $dv = $_[0];

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	$dbh->do("DELETE FROM  vlans where device = \"$dv\"");
	my $sth = $dbh->prepare("INSERT INTO vlans(device,vlanid,vlanname) VALUES ( ?,?,? )");
	foreach my $i ( sort keys %{$main::vlan{$dv}} ){
		$sth->execute ( $dv,$i,$main::vlan{$dv}{$i} );
		$nvlans++;
	}
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
	print "VLAN:	$nvlans vlans written to $misc::dbname.vlans\n" if $main::opt{v};
}

#===================================================================
# Write the network table
# Parameters:	devicename
# Global:	-
# Return:	-
#===================================================================
sub WriteNet {

	my $nnet = 0;

	my $dv = $_[0];

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	$dbh->do("DELETE FROM  networks where device = \"$dv\"");
	my $sth = $dbh->prepare("INSERT INTO networks(	device,ifname,ip,mask,vrfname,status ) VALUES ( ?,?,?,?,?,? )");
	foreach my $n ( sort keys %{$main::net{$dv}} ){
		$sth->execute (	$dv,
						$main::net{$dv}{$n}{ifn},
						&misc::Ip2Dec($n),
						&misc::Ip2Dec($main::net{$dv}{$n}{msk}),
						$main::net{$dv}{$n}{vrf},
						$main::net{$dv}{$n}{vrs} );
		$nnet++;
	}
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
	print "NET:	$nnet networks written to $misc::dbname.networks\n" if $main::opt{v};
}

#===================================================================
# Write the link table
# Parameters:	devicename
# Global:	-
# Return:	-
#===================================================================
sub WriteLink {

	my $nlink  = 0;
	my $nslink = 0;

	my $dv = $_[0];

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	$dbh->do("DELETE FROM links where device = \"$dv\" AND type != \"STAT\"");
	my $sth = $dbh->prepare("INSERT INTO links(device,ifname,neighbour,nbrifname,bandwidth,type,power,nbrduplex,nbrvlanid) VALUES ( ?,?,?,?,?,?,?,?,? )");

	foreach my $i ( sort keys %{$main::link{$dv}} ){
		foreach my $ne ( sort keys %{$main::link{$dv}{$i}} ){
			foreach my $ni ( sort keys %{$main::link{$dv}{$i}{$ne}} ){
				if($main::link{$dv}{$i}{$ne}{$ni}{ty} ne 'STAT'){
					if(!defined $main::link{$dv}{$i}{$ne}{$ni}{pr}){$main::link{$dv}{$i}{$ne}{$ni}{pr} = 0}
					$sth->execute (	$dv,$i,$ne,$ni,
							$main::link{$dv}{$i}{$ne}{$ni}{bw},
							$main::link{$dv}{$i}{$ne}{$ni}{ty},
							$main::link{$dv}{$i}{$ne}{$ni}{pr},
							$main::link{$dv}{$i}{$ne}{$ni}{du},
							$main::link{$dv}{$i}{$ne}{$ni}{vl} );
					$nlink++;
				}else{
					$nslink++;
				}
			}
		}
	}
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
	print "LINK:	$nlink (ignoring $nslink static) links written to $misc::dbname.links\n" if $main::opt{v};
}

#===================================================================
# Update Devices/Modules in Stock, which are discovered on the network.
# Parameters:	devicename
# Global:	-
# Return:	-
#===================================================================
sub UnStock {

	my $dv = $_[0];
	my $now = localtime($main::now);

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	if( $dbh->do("UPDATE stock SET location=\"Corporate\",comment=\"Discovered on $now as $dv\",state=100 where serial = \"$main::dev{$dv}{sn}\" and state !=100") + 0){
		print "\nStock: Discovered device $main::dev{$dv}{sn} set active in stock.\n" if $main::opt{v};
	}
	foreach my $i ( sort keys %{$main::mod{$dv}} ){
		if( $dbh->do("UPDATE stock SET location=\"Corporate\",comment=\"Discovered on $now in $dv slot $main::mod{$dv}{$i}{sl}\",state=100 where serial = \"$main::mod{$dv}{$i}{sn}\" and state !=100") + 0){
			print "\nStock: Discovered module $main::mod{$dv}{$i}{sn} set active in stock.\n" if $main::opt{v};
		}
	}
	$dbh->disconnect;
}

#===================================================================
# Write the nodes table
# Parameters:	-
# Global:	-
# Return:	-
#===================================================================
sub WriteNod {

	my $nnod = 0;

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	my $sth = $dbh->prepare("SELECT * FROM stolen");
	$sth->execute();
	my %stomac = ();
	while ((my @smac) = $sth->fetchrow_array) {
		$stomac{$smac[2]} = "$smac[6]";
	}
	$dbh->do("TRUNCATE nodes");
	$sth = $dbh->prepare("INSERT INTO nodes(	name,ip,mac,oui,
							firstseen,lastseen,
							device,ifname,vlanid,
							ifmetric,ifupdate,ifchanges,
							ipupdate,ipchanges,iplost,arp,tcpports,udpports,os,type,osupdate) VALUES ( ?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,? )");

	foreach my $mc ( sort keys %main::nod ){
		if (exists $stomac{$mc} and $main::nod{$mc}{ls} == $main::now and $misc::notify =~ /n/i){
			if( ! &Insert('messages','level,time,source,info',"\"150\",\"$main::now\",\"$mc\",\"Node has reappeared!\"") ){
				die "DB error messages!\n";
			}
			if($misc::notify =~ /N/){
				print &mon::SendMail("Stolen Node Warning!","Node $mc has reappeared with $main::nod{$mc}{ip} on $main::nod{$mc}{dv} $main::nod{$mc}{if}!");
			}
		}
		if(!defined $main::nod{$mc}{na}){$main::nod{$mc}{na}	= "-"}
		if(!defined $main::nod{$mc}{ip}){$main::nod{$mc}{ip}	= 0}
		if(!defined $main::nod{$mc}{an}){$main::nod{$mc}{an}	= 0}
		if(!defined $main::nod{$mc}{ou}){$main::nod{$mc}{ou}	= 0}
		if($main::nod{$mc}{vl} !~ /[0-9]+/){$main::nod{$mc}{vl}	= 0}

		my $dip = &misc::Ip2Dec($main::nod{$mc}{ip});

		$sth->execute (	$main::nod{$mc}{na},
				$dip,
				$mc,
				$main::nod{$mc}{nv},
				$main::nod{$mc}{fs},
				$main::nod{$mc}{ls},
				$main::nod{$mc}{dv},
				$main::nod{$mc}{if},
				$main::nod{$mc}{vl},
				$main::nod{$mc}{im},
				$main::nod{$mc}{iu},
				$main::nod{$mc}{ic},
				$main::nod{$mc}{au},
				$main::nod{$mc}{ac},
				$main::nod{$mc}{al},
				$main::nod{$mc}{an},
				$main::nod{$mc}{tp},
				$main::nod{$mc}{up},
				$main::nod{$mc}{os},
				$main::nod{$mc}{ty},
				$main::nod{$mc}{ou} );
		$nnod++;
	}
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;
	print "Node:	$nnod nodes written to $misc::dbname.nodes\n";
}

#===================================================================
# Delete RRDs & Configs of deleted devices
# Parameters:	-
# Global:	-
# Return:	-
#===================================================================
sub DelDev {

	my $nrrd = 0;
	my $ncfg = 0;

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $sth = $dbh->prepare("SELECT * FROM devdel");
	$sth->execute();
	my %devdel = ();
	while (my @dd = $sth->fetchrow_array) {
		$devdel{$dd[0]} = "$dd[1] - ". localtime($dd[2]);
	}
	$dbh->do("TRUNCATE devdel");
	$sth->finish if $sth;
	$dbh->disconnect;

	foreach my $na ( keys %devdel ){
		my $nrrd = &main::rmtree( "$misc::nedipath/rrd/$na");
		my $ncfg = &main::rmtree( "$misc::nedipath/conf/$na");
		print "DEL:	$nrrd RRD & $ncfg Config files (with dirs) deleted of $na for $devdel{$na}\n";
	}
}

#===================================================================
# Update Top traffic, error, power & monitoring RRD
# Parameters:	-
# Global:	-
# Return:	-
#===================================================================
sub TopRRD {
	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	# Using delta octets to avoid error from missing or rebooted switches. Needs to be divided by rrdstep*1M to get MB/s
	my $sth = $dbh->prepare("SELECT sum(dinoct)/(1000000*$misc::rrdstep),sum(doutoct)/(1000000*$misc::rrdstep) FROM interfaces WHERE linktype IS NULL");
	$sth->execute();
	(my @tet) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Wired interface (type not 71) errors/s
	my $sth = $dbh->prepare("SELECT sum(dinerr)/$misc::rrdstep,sum(douterr)/$misc::rrdstep FROM interfaces WHERE type != 71");
	$sth->execute();
	(my @twe) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Total Nodes lastseen
	my $sth = $dbh->prepare("SELECT count(*) FROM nodes where lastseen = $main::now");
	$sth->execute();
	(my @nodl) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Total Nodes firstseen
	my $sth = $dbh->prepare("SELECT count(*) FROM nodes where firstseen = $main::now");
	$sth->execute();
	(my @nodf) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Total power in Watts
	my $sth = $dbh->prepare("SELECT sum(power)/1000 FROM links");
	$sth->execute();
	(my @pwr) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Count all interfaces set administratively down during discovery cycle (having 8th bit set).
	my $sth = $dbh->prepare("SELECT count(*)  from interfaces WHERE status & 128");
	$sth->execute();
	(my @dis) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Number of monitored devices and last check
	my $sth = $dbh->prepare("select count(*),lastchk from monitoring where status = 0 group by lastchk order by lastchk desc");
	$sth->execute();
	(my @mok) = $sth->fetchrow_array;
	$sth->finish if $sth;

	if($mok[1] < (time - 1.5 * $misc::pause) ){						# moni.pl not running or no one alive
		$mok[0] = 0;
	}
	# Number of dead devices
	my $sth = $dbh->prepare("select count(*) from monitoring where status > 0");
	$sth->execute();
	(my @mal) = $sth->fetchrow_array;
	$sth->finish if $sth;
	# Number of cathegorized messages during discovery cycle
	my $sth = $dbh->prepare("SELECT level,count(*) FROM messages WHERE time > ".(time - $misc::rrdstep)." GROUP BY level");
	$sth->execute();
	while ((my @m) = $sth->fetchrow_array) {
		if($m[0] eq 50){
			$m50 = $m[1];
		}elsif($m[0] eq 100){
			$m100 = $m[1];
		}elsif($m[0] eq 150){
			$m150 = $m[1];
		}elsif($m[0] eq 200){
			$m200 = $m[1];
		}elsif($m[0] eq 250){
			$m250 = $m[1];
		}
	}
	$sth->finish if $sth;

	$dbh->disconnect;

	if ($main::opt{d}){
		print "RRD:\t$misc::nedipath/rrd/top.rrd Trf=$tet[0]/$tet[1] Err=$twe[0]/$twe[1] Nod=$nodl[0]/$nodf[0]\n";
		print "\tPwr=$pwr[0] Dif=$dis[0] Mon=$mok[0]/$mal[0] MSG=$m50,$m100,$m150,$m200,$m250";
	}
	if ($main::opt{t} or $main::opt{a}){
		print " No Write\n" if ($main::opt{d});
	}else{
		if (-e "$misc::nedipath/rrd/top.rrd"){
			$ok = 1;
		}else{
			my $ds = 2 * $misc::rrdstep;
			$ok = 1 + system ($misc::rrdcmd,
					"create","$misc::nedipath/rrd/top.rrd",
					"-s","$misc::rrdstep",
					"DS:tinoct:GAUGE:$ds:0:U",
					"DS:totoct:GAUGE:$ds:0:U",
					"DS:tinerr:GAUGE:$ds:0:U",
					"DS:toterr:GAUGE:$ds:0:U",
					"DS:tnodls:GAUGE:$ds:0:U",
					"DS:tnodfs:GAUGE:$ds:0:U",
					"DS:tpower:GAUGE:$ds:0:U",
					"DS:tdisif:GAUGE:$ds:0:U",
					"DS:monok:GAUGE:$ds:0:U",
					"DS:monal:GAUGE:$ds:0:U",
					"DS:msg50:GAUGE:$ds:0:U",
					"DS:msg100:GAUGE:$ds:0:U",
					"DS:msg150:GAUGE:$ds:0:U",
					"DS:msg200:GAUGE:$ds:0:U",
					"DS:msg250:GAUGE:$ds:0:U",
					"RRA:AVERAGE:0.5:1:2000",
					"RRA:AVERAGE:0.5:10:2000");
		}
		if($ok){
			if(1 + system ($misc::rrdcmd,"update","$misc::nedipath/rrd/top.rrd","N:$tet[0]:$tet[1]:$twe[0]:$twe[1]:$nodl[0]:$nodf[0]:$pwr[0]:$dis[0]:$mok[0]:$mal[0]:$m50:$m100:$m150:$m200:$m250") ){
				print " OK\n" if ($main::opt{d});
			}else{
				print " Update!\n" if ($main::opt{d});
			}
		}else{
			print " File!\n" if ($main::opt{d});
		}
	}
}

#===================================================================
# Update WLAN table.
# Parameters:	-
# Global:	-
# Return:	-
#===================================================================
sub WlanUp {

	use File::Find;

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 0});
	my $sth = $dbh->prepare("SELECT * FROM wlan");
	$sth->execute();

	while ((my @wrow) = $sth->fetchrow_array) {
			my $mc = $wrow[0];
			$ap{$mc} = $main::now;
	}
	my $wprev = keys %ap;
	print "$wprev	old Wlan entries read.\n";

	find(\&misc::GetAp, $main::opt{w});							# Calls GetAp() in libmisc.pl

	$dbh->do("TRUNCATE wlan");
	$sth = $dbh->prepare("INSERT INTO wlan(mac,time) VALUES ( ?,? )");
	for my $mc (sort keys %ap ){ $sth->execute ( $mc,$ap{$mc} ) }
	$dbh->commit;
	$sth->finish if $sth;
	$dbh->disconnect;

	my $wnew = scalar keys %ap;
	print "Wlan:	$wnew new entries written.\n";
}

#===================================================================
# Read Monitor table.
# Parameters:	(optional column, values to match)
# Global:	-
# Return:	-
#===================================================================
sub ReadMon {

	my $nmon  = 0;
	my $where = "";

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	if($_[0] and $_[1]){$where = "WHERE $_[0] = \"$_[1]\""}
	my $sth = $dbh->prepare("SELECT * FROM monitoring $where");
	$sth->execute();
	if($sth->rows){
		undef (%main::mon);								# Replace entries only when we got some new ones. Avoid problems in moni.pl while table is being updated
	}
	while ((my @f) = $sth->fetchrow_array) {
		my $d = $f[0];
		$main::mon{$d}{st} = $f[1];
		$main::mon{$d}{dp} = $f[2];
		$main::mon{$d}{ss} = $f[3];
		$main::mon{$d}{ml} = $f[4];
		$main::mon{$d}{lc} = $f[5];
		$main::mon{$d}{ut} = $f[6];
		$main::mon{$d}{lt} = $f[7];
		$main::mon{$d}{ok} = $f[8];
		$nmon++;
	}
	$sth->finish if $sth;
	$dbh->disconnect;
	print "Monitor:	$nmon entries ($where) read from $misc::dbname.monitoring\n" if (!$main::opt{D});
}

#===================================================================
# Read User table.
# Parameters:	(optional column, values to match)
# Global:	-
# Return:	-
#===================================================================
sub ReadUser {

	my $nusr  = 0;
	my $where = "";
	if($_[0]){$where = "WHERE groups & $_[0]"}

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $sth = $dbh->prepare("SELECT * FROM user $where");
	$sth->execute();
	while ((my @f) = $sth->fetchrow_array) {
		my $u = $f[0];
		$main::usr{$u}{ml} = $f[3];
		$main::usr{$u}{ph} = $f[4];
		$nusr++;
	}
	$sth->finish if $sth;
	$dbh->disconnect;
	print "User:	$nusr entries ($where) read from $misc::dbname.user\n" if (!$main::opt{D});	# Don't print if daemonized in sylog and moni
}

#===================================================================
# Update DB value(s)
# Parameters:	table, reference to hash of to be set columns, match expression
# Global:	-
# Return:	result
#===================================================================
sub Update {

	my @setq = ();
	my $table = shift (@_);
	my %scol = %{shift (@_)};
	my $match = shift (@_);

	foreach my $s ( keys %scol ){
		push(@setq,"$s=\"$scol{$s}\"");
	}
	my $set = join(',',@setq);
	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $r = $dbh->do("UPDATE $table SET $set WHERE $match");
	$dbh->disconnect;
	return $r;
}

#===================================================================
# Insert DB Record
# Parameters:	table, string of columns, string of values
# Global:	-
# Return:	-
#===================================================================
sub Insert {

	my $nag_event_service = 'Events';

	if ($misc::nagpipe and $_[0] eq 'messages') {
		if (-p $misc::nagpipe) {							# Nagios Handler by S.Neuser
			my ($level_str,$time,$source,$msg) = split /,/, $_[2];
			$level_str =~ s/\"//g;
			my $level = int $level_str;
			my $status = 3;
			if (! defined $level) { $status = 3; }
			elsif ($level < 0) { $status = 3; }					# UNKNOWN
			elsif ($level < 100) { $status = 0; }					# OK
			elsif ($level < 200) { $status = 1; }					# WARN
			else { $status = 2; }							# CRIT
			my $lsource = lc ($source);
			$lsource =~ s/\"//g;
			$time =~ s/\"//g;
			$msg =~ s/\"//g;
			$msg =~ s/\n/;/g;
			open (NPIPE, ">>$misc::nagpipe");
			print NPIPE "[$time] PROCESS_SERVICE_CHECK_RESULT;$lsource;$nag_event_service;$status;NeDi:$msg\n";
			close NPIPE;
		}
	}
	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $r = $dbh->do("insert into $_[0] ($_[1]) values ($_[2])");
	$dbh->disconnect;
	return $r;
}

#===================================================================
# Delete DB Record
# Parameters:	table, column, value
# Global:	-
# Return:	-
#===================================================================
sub Delete {

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $r = $dbh->do("delete from  $_[0] where $_[1] = \"$_[2]\"");
	$dbh->disconnect;
	return $r;
}

#===================================================================
# Set a value in the system table
# Parameters:	name, value
# Global:	-
# Return:	-
#===================================================================
sub SetSystem {

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $r = $dbh->do("UPDATE system SET value=$_[1] WHERE name=\"$_[0]\"");
	$dbh->disconnect;
	return $r;
}

#===================================================================
# Get a value in the system table
# Parameters:	name
# Global:	-
# Return:	value
#===================================================================
sub GetSystem {

	my $dbh = DBI->connect("DBI:mysql:$misc::dbname:$misc::dbhost", "$misc::dbuser", "$misc::dbpass", { RaiseError => 1, AutoCommit => 1});
	my $sth = $dbh->prepare("SELECT value FROM system WHERE name=\"$_[0]\"");
	$sth->execute();
	my @value = $sth->fetchrow_array;
	$sth->finish if $sth;
	$dbh->disconnect;
	return $value[0];
}

1;
