#============================================================================
#
# Program: libsnmp.pl
# Programmer: Remo Rickli
#
# -> SNMP based Functions <-
#
#============================================================================
package snmp;
use Net::SNMP;

#===================================================================
# Find community and identify device based on sysobj definition
# Parameters:	IP address
# Global:	misc::sysobj, main::dev
# Return:	name on success, empty string on failure
#===================================================================
sub Identify {

	my $comm	= "";
	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= 0;
	my $name 	= "";

	my $sysO	= '1.3.6.1.2.1.1.2.0';
	my $conO	= '1.3.6.1.2.1.1.4.0';
	my $namO	= '1.3.6.1.2.1.1.5.0';
	my $locO	= '1.3.6.1.2.1.1.6.0';
	my $srvO	= '1.3.6.1.2.1.1.7.0';

	my $peer	= $misc::doip{$_[0]};
	my $snmpver	= 2;									# Will be overwritten with snmpini, if set.

	print "$peer\t";
	if($peer !~ /$misc::netfilter/){
		if($misc::notify =~ /d/){
			if( ! &db::Insert('messages','level,time,source,info',"\"100\",\"$main::now\",\"$peer\",\"Netfilter $misc::netfilter, not discoverable\"") ){
				die "DB error messages!\n";
			}
		}
		print "Netfilter $misc::netfilter\t";
		return "";
	}
	my @comms = @misc::comms;

	if ($misc::snmpini{$peer}{cm}){								# Build Community list, respecting seedlist/db data.
		unshift(@comms,$misc::snmpini{$peer}{cm});
	}
	do {
		$comm = shift (@comms);
		$snmpver	= $misc::snmpini{$peer}{sp} if $misc::snmpini{$peer}{sp};
		print " C:$comm" if $main::opt{d};
		if($misc::comms{$comm}{aprot}){
			$snmpver = 3;
			print "-v$snmpver" if $main::opt{d};
			($session, $error) = Net::SNMP->session(-hostname  => $peer,
								-timeout   => $misc::timeout,
								-version   => $snmpver,
								-translate => [-octetstring => 0x0]
								-username      => $comm,
								-authpassword  => $misc::comms{$comm}{apass},
								-authprotocol  => $misc::comms{$comm}{aprot},
								-privpassword  => $misc::comms{$comm}{ppass},
								-privprotocol  => $misc::comms{$comm}{pprot}
								);
		}else{
			print "-v$snmpver" if $main::opt{d};
			($session, $error) = Net::SNMP->session(-hostname  => $peer,
								-community => $comm,
								-timeout   => $misc::timeout,
								-version   => $snmpver,
								-translate => [-octetstring => 0x0]
								);
		}
		$r = $session->get_request($namO);						# Get name to find the right community
		$err = $session->error;
		$session->close;
		if($err and $snmpver eq 2){							# Fall back to version 1 if 2 failed
			$snmpver = 1;
			print "-v$snmpver" if $main::opt{d};
			($session, $error) = Net::SNMP->session(-hostname  => $peer,
								-community => $comm,
								-timeout   => $misc::timeout,
								-version   => $snmpver,
								-translate => [-octetstring => 0x0]
								);

			$r = $session->get_request($namO);					# Get name to find the right community
			$err = $session->error;
			$session->close;
		}
	} while ($#comms ne "-1" and $err);							# And stop once a community worked or we ran out of them.
	if ($err){
		if($misc::notify =~ /d/){
			if( ! &db::Insert('messages','level,time,source,info',"\"100\",\"$main::now\",\"$peer\",\"$_[0] no name, not discoverable\"") ){
				die "DB error messages!\n";
			}
		}
		print "No name, not discoverable\t\t";
	}else{
		print "+ " if $main::opt{d};
		$name = &misc::Strip($r->{$namO});
		if ($name =~ /^\s*$/){
			$name = $peer;
		}else{
			$name =~  s/^(.*?)\.(.*)/$1/ if !$main::opt{F};				# FQDN can mess up links
		}
		if($main::opt{d} or $main::opt{v}){
			print "$name\t";
		}else{
			printf ("%-12.12s\t",$name);
		}
		if(grep /^\Q$name\E$/, @misc::donenam){						# Bail, if we've done it already
			print "Done already\t\t\t";
			return "";
		}
		($session, $error) = Net::SNMP->session(-hostname  => $peer,
							-community => $comm,
							-timeout   => $misc::timeout,
							-version   => $snmpver,
							-translate => [-octetstring => 0x0]
							);

		my $so	= "other";
		$r	= $session->get_request($sysO);
		$err	= $session->error;
		if(!$err and defined $r->{$sysO}){$so = &misc::Strip($r->{$sysO})}

		if(!exists $misc::sysobj{$so}){							# Load .def if not done already
			if (-e "$main::p/sysobj/$so.def"){
				open  ("DEF", "$main::p/sysobj/$so.def");
			}else{
				open  ("DEF","$main::p/sysobj/other.def");
			}
			my @def = <DEF>;
			close("DEF");
			chomp @def;
			$misc::sysobj{$so}{bf} = "";
			$misc::sysobj{$so}{ty} = $so;
			$misc::sysobj{$so}{hc} = 0;
			$misc::sysobj{$so}{mv} = 0;
			foreach my $l (@def){
				if ($l !~ /^[#;]|^\s*$/){
					my @v  = split(/\t+/,$l);
					if (!defined $v[1]){$v[1] = ""}
					if ($v[0] eq "Type")		{$misc::sysobj{$so}{ty} = $v[1]}
					elsif ($v[0] eq "OS")		{$misc::sysobj{$so}{os} = $v[1]}
					elsif ($v[0] eq "Icon")		{$misc::sysobj{$so}{ic} = $v[1]}
					elsif ($v[0] eq "SNMPv")	{
						$misc::sysobj{$so}{sp} = substr($v[1],0,1);
						if(substr($v[1],1,2) eq 'HC'){$misc::sysobj{$so}{hc} = 128}
					}
					elsif ($v[0] eq "Serial")	{$misc::sysobj{$so}{sn} = $v[1]}
					elsif ($v[0] eq "Bimage")	{$misc::sysobj{$so}{bi} = $v[1]}
					elsif ($v[0] eq "Sysdes")	{$misc::sysobj{$so}{de} = $v[1]}
					elsif ($v[0] eq "Bridge")	{$misc::sysobj{$so}{bf} = $v[1]}
					elsif ($v[0] eq "Dispro")	{$misc::sysobj{$so}{dp} = $v[1]}

					elsif ($v[0] eq "VLnams")	{$misc::sysobj{$so}{vn} = $v[1]}
					elsif ($v[0] eq "VTPdom")	{$misc::sysobj{$so}{vd} = $v[1]}
					elsif ($v[0] eq "VTPmod")	{$misc::sysobj{$so}{vm} = $v[1]}

					elsif ($v[0] eq "IFalia")	{$misc::sysobj{$so}{al} = $v[1]}
					elsif ($v[0] eq "IFalix")	{$misc::sysobj{$so}{ax} = $v[1]}
					elsif ($v[0] eq "IFdupl")	{$misc::sysobj{$so}{du} = $v[1]}
					elsif ($v[0] eq "IFduix")	{$misc::sysobj{$so}{dx} = $v[1]}
					elsif ($v[0] eq "Halfdp")	{$misc::sysobj{$so}{hd} = $v[1]}
					elsif ($v[0] eq "Fulldp")	{$misc::sysobj{$so}{fd} = $v[1]}
					elsif ($v[0] eq "IFvlan")	{$misc::sysobj{$so}{vi} = $v[1]}
					elsif ($v[0] eq "IFvlix")	{$misc::sysobj{$so}{vx} = $v[1]}

					elsif ($v[0] eq "Modesc")	{$misc::sysobj{$so}{md} = $v[1]}
					elsif ($v[0] eq "Moclas")	{$misc::sysobj{$so}{mc} = $v[1]}
					elsif ($v[0] eq "Movalu")	{$misc::sysobj{$so}{mv} = $v[1]}
					elsif ($v[0] eq "Mostep")	{$misc::sysobj{$so}{mp} = $v[1]}
					elsif ($v[0] eq "Moslot")	{$misc::sysobj{$so}{mt} = $v[1]}
					elsif ($v[0] eq "Modhw")	{$misc::sysobj{$so}{mh} = $v[1]}
					elsif ($v[0] eq "Modsw")	{$misc::sysobj{$so}{ms} = $v[1]}
					elsif ($v[0] eq "Modfw")	{$misc::sysobj{$so}{mf} = $v[1]}
					elsif ($v[0] eq "Modser")	{$misc::sysobj{$so}{mn} = $v[1]}
					elsif ($v[0] eq "Momodl")	{$misc::sysobj{$so}{mm} = $v[1]}


					elsif ($v[0] eq "CPUutl")	{$misc::sysobj{$so}{rc} = $v[1]}
					elsif ($v[0] eq "MemCPU")	{$misc::sysobj{$so}{rm} = $v[1]}
					elsif ($v[0] eq "Temp")		{$misc::sysobj{$so}{rt} = $v[1]}
					elsif ($v[0] eq "MemIO")	{$misc::sysobj{$so}{rv} = $v[1];$misc::sysobj{$so}{rl} = "Mem IO;G;Bytes free"}
					elsif ($v[0] eq "Custom")	{$misc::sysobj{$so}{rv} = $v[2];$misc::sysobj{$so}{rl} = $v[1]}
				}
			}
		}
		my $desO = ($misc::sysobj{$so}{de})?$misc::sysobj{$so}{de}:'1.3.6.1.2.1.1.1.0';	# Use sysdesc OID from .def, if set
		$r = $session->get_request( $desO );
		$err = $session->error;
		my $de = "err";
		if(!$err){$de = &misc::Strip($r->{$desO})}
		if (defined $misc::descfilter and $de =~ /$misc::descfilter/){			# Only define device, if not filtered in nedi.conf
			$session->close;
			if($misc::notify =~ /d/){
				if( ! &db::Insert('messages','level,time,source,info',"\"100\",\"$main::now\",\"$peer\",\"Descfilter $misc::descfilter, not discoverable\"") ){
					die "DB error messages!\n";
				}
			}
			print "Descfilter $misc::descfilter\t\t";
			return "";
		}else{
			$main::dev{$name}{so} = $so;
			$main::dev{$name}{ty} = $misc::sysobj{$so}{ty};
			$main::dev{$name}{ip} = $peer;
			$main::dev{$name}{cm} = $comm;
			$main::dev{$name}{de} = $de;
			$main::dev{$name}{os} = $misc::sysobj{$so}{os};
			$main::dev{$name}{ic} = $misc::sysobj{$so}{ic};
			$main::dev{$name}{sp} = $misc::sysobj{$so}{sp};
			$main::dev{$name}{hc} = (defined $main::dev{$name}{hc})?$main::dev{$name}{hc}:$misc::sysobj{$so}{hc};
			$r = $session->get_request($conO);
			$err = $session->error;
			if(!$err){$main::dev{$name}{co} = &misc::Strip($r->{$conO})}
			$r = $session->get_request($locO);
			$err = $session->error;
			if(!$err){$main::dev{$name}{lo} = &misc::Strip($r->{$locO})}
			$r = $session->get_request($srvO);
			$err = $session->error;
			if($err or $r->{$srvO} !~ /^\d+$/){
				$main::dev{$name}{sv} = 6; 					# Could be a buggy SNMP implementation, so we set this to 6 and check the device anyway
			}else{
				$main::dev{$name}{sv} = &misc::Strip($r->{$srvO});
			}
			print "SV=$main::dev{$name}{sv} TY=$main::dev{$name}{ty} $de" if $main::opt{d};
		}
		$session->close;
	}
	return $name;
}

#===================================================================
# Get enterprise specific information using sysobj.def file
# Parameters:	device name
# Global:	main::dev
# Return:	-
#===================================================================
sub Enterprise {

	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";
	my $nv		= 0;
	my $so		= $main::dev{$_[0]}{so};

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
						-community => $main::dev{$_[0]}{cm},
						-timeout   => $misc::timeout,
						-version   => $main::dev{$_[0]}{sp},
						-translate => [-octetstring => 0x0]
						);

	if($misc::sysobj{$so}{sn}){
		$r  = $session->get_request($misc::sysobj{$so}{sn});
		$err = $session->error;
		if ($err){
			print "Sn";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{sn} = "err";
		}else{
			$main::dev{$_[0]}{sn} = &misc::Strip( substr($r->{$misc::sysobj{$so}{sn}},0,31) );
		}
	}else{
		$main::dev{$_[0]}{sn} = "-";
	}
	if($misc::sysobj{$so}{bi}){
		$r  = $session->get_request($misc::sysobj{$so}{bi});
		$err = $session->error;
		if ($err){
			print "Sb";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{bi} = "err";
		}else{
			my $bimg = &misc::Strip($r->{$misc::sysobj{$so}{bi}});
			$bimg =~ s/^flash:|^bootflash:|^slot[0-9]:|^sup-boot(flash|disk):|^disk0:|FIRMWARE REVISION: //;
			$bimg =~ s/.*\/(.*)/$1/;
			$main::dev{$_[0]}{bi} = substr($bimg,0,63);
			print "BI=$main::dev{$_[0]}{bi} " if $main::opt{d};
		}
	}else{
		$main::dev{$_[0]}{bi} = "-";
	}
	if($misc::sysobj{$so}{rc}){
		$r  = $session->get_request($misc::sysobj{$so}{rc});
		$err = $session->error;
		if ($err or $r->{$misc::sysobj{$so}{rc}} !~ /^[0-9]+$/){
			print "Su";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{cpu} = 0;
		}else{
			$main::dev{$_[0]}{cpu} = &misc::Strip($r->{$misc::sysobj{$so}{rc}});
			if($misc::notify =~ /s/ and $main::dev{$_[0]}{cpu} > $misc::cpua){
				if( ! &db::Insert('messages','level,time,source,info',"\"200\",\"$main::now\",\"$_[0]\",\"CPU load of $main::dev{$_[0]}{cpu}% detected!\"") ){
					die "DB error messages!\n";
				}
			}
		}
		print "CPU=$main::dev{$_[0]}{cpu} " if $main::opt{d};
	}else{
		$main::dev{$_[0]}{cpu} = 0;
	}
	if($misc::sysobj{$so}{rm}){
		$r  = $session->get_request($misc::sysobj{$so}{rm});
		$err = $session->error;
		if ($err or $r->{$misc::sysobj{$so}{rm}} !~ /^[0-9]+$/){
			print "Sm";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{mcp} = 0;
		}else{
			$main::dev{$_[0]}{mcp} = &misc::Strip($r->{$misc::sysobj{$so}{rm}});
			if($misc::notify =~ /s/ and $main::dev{$_[0]}{mcp} < $misc::mema){
				if( ! &db::Insert('messages','level,time,source,info',"\"200\",\"$main::now\",\"$_[0]\",\"CPU memory of $main::dev{$_[0]}{mcp} bytes detected!\"") ){
					die "DB error messages!\n";
				}
			}
		}
		print "Mem=$main::dev{$_[0]}{mcp} " if $main::opt{d};
	}else{
		$main::dev{$_[0]}{mcp} = 0;
	}
	if($misc::sysobj{$so}{rv}){
		$r  = $session->get_request($misc::sysobj{$so}{rv});
		$err = $session->error;
		if ($err or $r->{$misc::sysobj{$so}{rv}} !~ /^[0-9]+$/){
			print "Si";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{cuv} = 0;
		}else{
			$main::dev{$_[0]}{cuv} = $r->{$misc::sysobj{$so}{rv}};
			$main::dev{$_[0]}{cul} = $misc::sysobj{$so}{rl};
		}
		print "$misc::sysobj{$so}{rl}=$main::dev{$_[0]}{cuv} " if $main::opt{d};
	}else{
		$main::dev{$_[0]}{cuv} = 0;
	}
	if($misc::sysobj{$so}{rt}){
		$r  = $session->get_request($misc::sysobj{$so}{rt});
		$err = $session->error;
		if ($err or $r->{$misc::sysobj{$so}{rt}} !~ /^[0-9]+$/){
			print "St";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{tmp} = 0;
		}else{
			$main::dev{$_[0]}{tmp} = &misc::Strip($r->{$misc::sysobj{$so}{rt}});
			if($misc::notify =~ /s/ and $main::dev{$_[0]}{tmp} > $misc::tmpa){
				if( ! &db::Insert('messages','level,time,source,info',"\"200\",\"$main::now\",\"$_[0]\",\"Temperature of  $main::dev{$_[0]}{tmp} degrees Celsius detected!\"") ){
					die "DB error messages!\n";
				}
			}

		}
		print "Temp=$main::dev{$_[0]}{tmp} " if $main::opt{d};
	}else{
		$main::dev{$_[0]}{tmp} = 0;
	}
	if($misc::sysobj{$so}{vd}){
		$r  = $session->get_request($misc::sysobj{$so}{vd});
		$err = $session->error;
		if ($err){
			print "Vd";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{vd} = "?";
		}else{
			$main::dev{$_[0]}{vd} = &misc::Strip($r->{$misc::sysobj{$so}{vd}});
		}
		print " VTP domain $main::dev{$_[0]}{vd} " if $main::opt{v};
	}else{
		$main::dev{$_[0]}{vd} = "-";
	}
	if($misc::sysobj{$so}{vm}){
		$r  = $session->get_request($misc::sysobj{$so}{vm});
		$err = $session->error;
		if ($err or $r->{$misc::sysobj{$so}{vm}} !~ /^[1-3]$/){
			print "Vm";
			print "$err\n" if $main::opt{d};
			$main::dev{$_[0]}{vm} = 9;
		}else{
			$main::dev{$_[0]}{vm} = $r->{$misc::sysobj{$so}{vm}};
		}
		print " mode $main::dev{$_[0]}{vm}\n" if $main::opt{v};
	}else{
		$main::dev{$_[0]}{vm} = 0;
	}
	if($misc::sysobj{$so}{vn}){
		$r = $session->get_table($misc::sysobj{$so}{vn});				# Get Vlan names
		$err = $session->error;
		if ($err){
			print "Vn";
			print "$err\n" if $main::opt{d};
		}else{
			while ( (my $vlO,my $na) =  each(%{$r}) ){
				my $vl = substr($vlO,rindex($vlO,'.') + 1);
				if($vl =~ /^[0-9]+$/){						# Use if vlanid is number!
					$main::vlan{$_[0]}{$vl} = $na;
				}else{
					print "Vi";
				}
				print "\n VL:$vl\t$na" if $main::opt{v};
				$nv++;
			}
			print "v$nv" if !$main::opt{v};
		}
	}
	$session->close;
}

#===================================================================
# Get interface information.
# Parameters:	device name
# Global:	main::dev, main::int, misc::ifmac, misc::portprop
# Return:	-
#===================================================================
sub Interfaces {

	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";
	my $io		= "";
	my $oo		= "";
	my $sp		= "";

	my $mergehc	= 0;
	my $warn	= 0;
	my $ni		= 0;

	my %ifde	= ();
	my %iftp	= ();
	my %ifsp	= ();
 	my %ifmc	= ();
	my %ifas	= ();

	my %ifio	= ();
	my %ifie	= ();
	my %ifoo	= ();
	my %ifoe	= ();
	my %ifna	= ();

	my %ifal	= ();
	my %ifax	= ();
	my %alias	= ();
	my %ifvl	= ();
	my %ifvx	= ();
	my %vlid	= ();

	my %ifbr	= ();
	my %ifdp	= ();
	my %ifdx	= ();
	my %duplex	= ();
	my %usedoid	= ();

	my $ifdesO	= '1.3.6.1.2.1.2.2.1.2';
	my $iftypO	= '1.3.6.1.2.1.2.2.1.3';
	my $ifspdO	= '1.3.6.1.2.1.2.2.1.5';
 	my $ifmacO	= '1.3.6.1.2.1.2.2.1.6';
	my $ifadmO	= '1.3.6.1.2.1.2.2.1.7';
	my $ifinoO	= '1.3.6.1.2.1.2.2.1.10';
	my $ifineO	= '1.3.6.1.2.1.2.2.1.14';
	my $ifotoO	= '1.3.6.1.2.1.2.2.1.16';
	my $ifoteO	= '1.3.6.1.2.1.2.2.1.20';

	my $ifnamO	= '1.3.6.1.2.1.31.1.1.1.1';
	my $ifibrO	= '1.3.6.1.2.1.31.1.1.1.3';
	my $ifhioO	= '1.3.6.1.2.1.31.1.1.1.6';
	my $ifhooO	= '1.3.6.1.2.1.31.1.1.1.10';
	my $ifhspO	= '1.3.6.1.2.1.31.1.1.1.15';

	my $ifaliO	= '1.3.6.1.2.1.31.1.1.1.18';

	my $dv = $_[0];

	if($misc::sysobj{$main::dev{$dv}{so}}{al}){
		$ifaliO = $misc::sysobj{$main::dev{$dv}{so}}{al};
	}
	my $ifalxO	= $misc::sysobj{$main::dev{$dv}{so}}{ax};
	my $ifvlaO	= $misc::sysobj{$main::dev{$dv}{so}}{vi};
	my $ifvlxO	= $misc::sysobj{$main::dev{$dv}{so}}{vx};
	my $ifdupO	= $misc::sysobj{$main::dev{$dv}{so}}{du};
	my $ifduxO	= $misc::sysobj{$main::dev{$dv}{so}}{dx};

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$dv}{ip},
						-community => $main::dev{$dv}{cm},
						-timeout   => $misc::timeout * 2,		# Big dev need more time (Vasily)
						-version   => $main::dev{$dv}{sp},
						-translate => [-octetstring => 0x0]
						);

	$session->max_msg_size(2048);								# Nortel Switches need this
	$r = $session->get_table($ifdesO);							# Walk interface description.
	$err = $session->error;
	if ($err){print "Id";print "$err\n" if $main::opt{d};$warn++}else{%ifde  = %{$r}}

	$r = $session->get_table($ifnamO);							# Walk interface name.
	$err = $session->error;
	if ($err){print "In";print "$err\n" if $main::opt{d};$warn++}else{ %ifna = %{$r}}

	$r = $session->get_table($iftypO);							# Walk interface type.
	$err = $session->error;
	if ($err){print "It";print "$err\n" if $main::opt{d};$warn++}else{%iftp  = %{$r}}

	$r = $session->get_table($ifmacO);							# Walk interface mac address.
	$err = $session->error;
	if ($err){print "Im";print "$err\n" if $main::opt{d};$warn++}else{%ifmc  = %{$r}}

	$r = $session->get_table($ifadmO);							# Walk interface admin status
	$err = $session->error;
	if ($err){print "Ia";print "$err\n" if $main::opt{d};$warn++}else{%ifas  = %{$r}}
	if($main::dev{$dv}{hc}){								# Walk interface HC if set
		$r = $session->get_table($ifhioO);
		$err = $session->error;
		if ($err){
			print "II";
			print "$err\n" if $main::opt{d};
			$main::dev{$dv}{hc} = 0;						# Remember non-HC support for future reference
			$warn++;
		}else{
			$io = $ifhioO;
			%ifio  = %{$r};
			$r = $session->get_table($ifhooO);					# Walk interface HC out octets
			$err = $session->error;
			if ($err){
				print "IO";
				print "$err\n" if $main::opt{d};
				$main::dev{$dv}{hc} = 0;					# Remember non-HC support for future reference
				$warn++
			}else{
				$oo = $ifhooO;
				%ifoo  = %{$r};
			}
			$r = $session->get_table($ifhspO);					# Walk interface high-speed
			$err = $session->error;
			if ($err){
				print "IS";
				print "$err\n" if $main::opt{d};
				$warn++
			}else{
				$sp = $ifhspO;
				%ifsp  = %{$r};
			}
			if( (scalar (keys %ifio) + 1) < scalar (keys %ifde) ){			# More than 1 missing HCctr > Vasily's RFC2233 fix
				$mergehc = 1;
				print " HC-Merge " if $main::opt{d};
			}
		}
	}
	if(!$main::dev{$dv}{hc} or $mergehc){
		$r = $session->get_table($ifinoO);						# Walk interface in octets otherwhise
		$err = $session->error;
		if ($err){
			print "Ii";
			print "$err\n" if $main::opt{d};
			$warn++;
		}else{
			if($mergehc){
				while( my($key, $val) = each(%{$r}) ) {
					my @if  = split(/\./,$key);
					my $i   = $if[10];
					$ifio{"$io.$i"} = $val if !$ifio{"$io.$i"};
				}
			}else{
				$io = $ifinoO;
				%ifio  = %{$r};
			}
			$r = $session->get_table($ifotoO);					# Walk interface out octets otherwhise
			$err = $session->error;
			if ($err){
				print "Io";
				print "$err\n" if $main::opt{d};
				$warn++;
			}else{
				if($mergehc){
					while( my($key, $val) = each(%{$r}) ) {
						my @if  = split(/\./,$key);
						my $i   = $if[10];
						$ifoo{"$oo.$i"} = $val if !$ifoo{"$oo.$i"};
					}
				}else{
					$oo = $ifotoO;
					%ifoo  = %{$r};
				}
			}
		}
		$r = $session->get_table($ifspdO);						# Walk interface speed.
		$err = $session->error;
		if ($err){
			print "Is";
			print "$err\n" if $main::opt{d};
			$warn++
		}else{
			if($mergehc){
				while( my($key, $val) = each(%{$r}) ) {
					my @if  = split(/\./,$key);
					my $i   = $if[10];
					$ifsp{"$sp.$i"} = $val/1000000 if !$ifsp{"$sp.$i"};
				}
			}else{
				$sp = $ifspdO;
				%ifsp  = %{$r};
			}
		}
	}
	$r = $session->get_table($ifineO);							# Walk interface in errors
	$err = $session->error;
	if ($err){print "Ie";print "$err\n" if $main::opt{d};$warn++}else{ %ifie  = %{$r}}

	$r = $session->get_table($ifoteO);							# Walk interface in errors
	$err = $session->error;
	if ($err){print "IE";print "$err\n" if $main::opt{d};$warn++}else{ %ifoe  = %{$r}}

	$r = $session->get_table($ifaliO);							# Walk interface alias.
	$err = $session->error;
	if ($err){print "Il";print "$err\n" if $main::opt{d};$warn++}else{ %ifal  = %{$r}}
	if($ifalxO){
		$r = $session->get_table($ifalxO);						# Walk index table if specified...
		$err = $session->error;
		if ($err){
			print "Il";
			print "$err\n" if $main::opt{d};
		}else{
			%ifax  = %{$r};
			$usedoid{$ifalxO} = \%ifax;						# (store in case it's the same for vlans or duplex)
			foreach my $x ( keys %ifax ){						# ...and map directly to if indexes
				my $i = $x;
				$i =~ s/$ifalxO//;
				$alias{$ifax{$x}} = $ifal{"$ifaliO$i"};
			}
		}
	}else{											# Else use indexes directly
		foreach my $x ( keys %ifal ){
			my $i = $x;
			$i =~ s/$ifaliO\.//;
			$alias{$i} = $ifal{$x};
		}
	}
	if($ifvlaO){										# Same for IF vlans...
		$r = $session->get_table($ifvlaO);
		$err = $session->error;
		if ($err){print "Iv";print "$err\n" if $main::opt{d};$warn++}else{ %ifvl  = %{$r}}
	}
	if($ifvlxO){										# If vlans use a different index
		if(exists $usedoid{$ifvlxO}){							# and if it's been used before
			%ifvx = %{$usedoid{$ifvlxO}};						# assign the vlan oid to where the used one points to.
		}else{										# Otherwhise walk it
			$r = $session->get_table($ifvlxO);
			$err = $session->error;
			if ($err){
				print "Iv";
				print "$err\n" if $main::opt{d};
			}else{
				%ifvx  = %{$r};
				$usedoid{$ifvlxO} = \%ifvx;
			}
		}
		foreach my $x ( keys %ifvx ){
			my $i = $x;
			$i =~ s/$ifvlxO\.//;
			$vlid{$ifvx{$x}} = $ifvl{"$ifvlaO.$i"};
		}
	}else{
		foreach my $x ( keys %ifvl ){
			my $i = $x;
			$i =~ s/$ifvlaO\.//;
			$vlid{$i} = $ifvl{$x};
		}
	}
	if($ifdupO){										# ...and IF duplex
		if($ifdupO eq "doublespeed"){							# If duplex is shown by speed...
			foreach my $x ( keys %ifsp ){
				my $i = $x;
				$i =~ s/$sp\.//;
				if($ifsp{$x} =~ /^20/){
					$ifsp{$x} /= 2;
					$duplex{$i} = "FD";
				}elsif($ifsp{$x} =~ /^10/){
					$duplex{$i} = "HD";
				}
			}
		}else{
			$r = $session->get_table($ifdupO);
			$err = $session->error;
			if ($err){print "Ix";print "$err\n" if $main::opt{d};$warn++}else{ %ifdp  = %{$r}}
		}
	}
	if($ifduxO){										# If duplex uses a different index
		if(exists $usedoid{$ifduxO}){							# and if it's been used before
			%ifdx = %{$usedoid{$ifduxO}};						# assign the duplex oid to where the used one points to.
		}else{										# Otherwhise walk it
			$r = $session->get_table($ifduxO);
			$err = $session->error;
			if ($err){
				print "Ix";
				print "$err\n" if $main::opt{d};
			}else{
				%ifdx  = %{$r};
				$usedoid{$ifduxO} = \%ifdx;
			}
		}
		foreach my $x ( keys %ifdx ){
			my $i = $x;
			$i =~ s/$ifduxO\.//;
			$duplex{$ifdx{$x}} = $ifdp{"$ifdupO.$i"};
		}
	}else{
		foreach my $x ( keys %ifdp ){
			my $i = $x;
			$i =~ s/$ifdupO\.//;
			$duplex{$i} = $ifdp{$x};
		}
	}
	if($main::dev{$dv}{fl} & 1){								# Dev flag #1 is set?
		$r = $session->get_table($ifibrO);						# Walk IF inbcasts
		$err = $session->error;
		if ($err){print "Ib";print "$err\n" if $main::opt{d};$warn++}else{ %ifbr  = %{$r}}
	}
	$session->close;

	foreach my $dek ( keys %ifde ){
		my @if  = split(/\./,$dek);
		my $i   = $if[10];								# ...is the IF index
		my $ina = $i;
		my $idn = (length($ifde{$dek}) < 30)?&misc::Shif("$ifde{$dek}"):"Port";
		if ($ifna{"$ifnamO.$i"} ){
			if(exists $misc::portprop{$dv}{$ifna{"$ifnamO.$i"}} ){
				$ina = $ifna{"$ifnamO.$i"} . "-$i";
			}else{
				$ina = $ifna{"$ifnamO.$i"};
			}
		}elsif( exists $misc::portprop{$dv}{$idn} ){
			$ina = "$idn-$i";							# Use IF desc and make unique, if necessary
		}else{
			$ina = $idn;								# Use IF desc if name is empty
		}
		$main::int{$dv}{$i}{ina} = $ina;
		$main::int{$dv}{$i}{lty} = "STAT" if(exists $main::link{$dv}{$ina});		# Set if static exists
		$main::int{$dv}{$i}{des} = (defined $ifde{$dek}?$ifde{$dek}:"");
		$main::int{$dv}{$i}{typ} = (defined $iftp{"$iftypO.$i"}?$iftp{"$iftypO.$i"}:0);
		$main::int{$dv}{$i}{spd} = (defined $ifsp{"$sp.$i"}?$ifsp{"$sp.$i"}:0);
		if($sp eq $ifhspO){$main::int{$dv}{$i}{spd} *= 1000000}
		$main::int{$dv}{$i}{sta} = (defined $ifas{"$ifadmO.$i"}?$ifas{"$ifadmO.$i"}:0);
		$main::int{$dv}{$i}{ioc} = (defined $ifio{"$io.$i"}?$ifio{"$io.$i"}:0);
		$main::int{$dv}{$i}{ooc} = (defined $ifoo{"$oo.$i"}?$ifoo{"$oo.$i"}:0);
		$main::int{$dv}{$i}{ier} = (defined $ifie{"$ifineO.$i"}?$ifie{"$ifineO.$i"}:0);
		$main::int{$dv}{$i}{oer} = (defined $ifoe{"$ifoteO.$i"}?$ifoe{"$ifoteO.$i"}:0);
		$main::int{$dv}{$i}{ali} = (defined $alias{$i}?$alias{$i}:"");
		$main::int{$dv}{$i}{vln} = (defined $vlid{$i}?$vlid{$i}:0);
		$main::int{$dv}{$i}{dio} = 0;
		$main::int{$dv}{$i}{die} = 0;
		$main::int{$dv}{$i}{doo} = 0;
		$main::int{$dv}{$i}{doe} = 0;
		$main::int{$dv}{$i}{com} = "";
		$main::int{$dv}{$i}{ibr} = (defined $ifbr{"$ifibrO.$i"}?$ifbr{"$ifibrO.$i"}:0);

		if($ifmc{"$ifmacO.$i"}){
			my $imac = unpack('H12', $ifmc{"$ifmacO.$i"});
			$main::int{$dv}{$i}{mac} = $imac;
			$misc::ifmac{$imac}++;
		}else{
			$main::int{$dv}{$i}{mac} = "";
		}
		if(defined $duplex{$i} ){							# Did we get a duplex value?
			if($duplex{$i} =~ /^[FH]D$/){						# Use if set properly already...
				$main::int{$dv}{$i}{dpx} = $duplex{$i};
			}else{									# ...or assign defined HD,FD key
				if($duplex{$i} eq $misc::sysobj{$main::dev{$_[0]}{so}}{fd}){$main::int{$dv}{$i}{dpx} = "FD"}
				elsif($duplex{$i} eq $misc::sysobj{$main::dev{$_[0]}{so}}{hd}){$main::int{$dv}{$i}{dpx} = "HD"}
				else{$main::int{$dv}{$i}{dpx} = "?"}
			}
		}else{
			$main::int{$dv}{$i}{dpx} = "-";
		}
		$misc::portprop{$dv}{$ina}{pop} = 0;
		$misc::portprop{$dv}{$ina}{idx} = $i;
		$misc::portprop{$dv}{$ina}{spd} = $main::int{$dv}{$i}{spd};
		$misc::portprop{$dv}{$ina}{vln} = $main::int{$dv}{$i}{vln};
		print "\n IF:$i\t$ina\tT:$main::int{$dv}{$i}{typ}\tD:$main::int{$dv}{$i}{dpx}\tVL:$main::int{$dv}{$i}{vln}\t$main::int{$dv}{$i}{spd}\t$main::int{$dv}{$i}{mac}\t$ifde{$dek}\t$main::int{$dv}{$i}{ali}" if $main::opt{v};
		$ni++;
	}
	print " i$ni" if !$main::opt{v};
	print ($warn?" ":"\t");
}

#===================================================================
# Get IP address tables and tries to find best mgmt IP (based on idea from Duane Walker)
# Parameters:	device name
# Global:	main::dev, main::net
# Return:	-
#===================================================================

sub IfAddresses {
	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";

	my %aifx	= ();
	my %ainm	= ();

	my %vrfs	= ();
	my %vrfna	= ();
	my %vrfst	= ();

	my $warn	= 0;
	my $nia		= 0;

	my $newip	= "";
	my $ippri	= 10;

	my $iaixO	= "1.3.6.1.2.1.4.20.1.2";
	my $ianmO	= "1.3.6.1.2.1.4.20.1.3";
	my $vrfO	= "1.3.6.1.3.118.1.2.1.1.6";

	my $dv = $_[0];

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$dv}{ip},
						-community => $main::dev{$dv}{cm},
						-timeout   => $misc::timeout,
						-version   => $main::dev{$dv}{sp},
						-translate => [-octetstring => 0x0]
						);

	$r   = $session->get_table($iaixO);
	$err = $session->error;
	if ($err){
		$session->close;
		print "Ji";
		print "$err\n" if $main::opt{d};
		return 1;
	}else{%aifx = %{$r}}

	$r   = $session->get_table($ianmO);
	$err = $session->error;
	if ($err){print "Jm";print "$err\n" if $main::opt{d};$warn++}else{%ainm = %{$r}}
	if ($main::dev{$dv}{sv} > 3 and $main::dev{$dv}{os} =~/IOS/){			# Try only Cisco routers for now
		$r   = $session->get_table($vrfO);
		$err = $session->error;
		if ($err){print "Jv";print "$err\n" if $main::opt{d};$warn++}else{ %vrfs = %{$r} }
		foreach my $k ( keys %vrfs ){
			my @karr = split(/\./,substr($k, 24));
			shift(@karr);
			my $ix = pop(@karr);							# Showstopper, no index shown in perl???
			my $vrfna = "";
			foreach my $char (@karr){						# VRF Name is OID...
				$vrfna .= chr($char);
			}
			$vrfna{$main::int{$dv}{$ix}{ina}} = $vrfna;
			$vrfst{$main::int{$dv}{$ix}{ina}} = $vrfs{$k};
		}
	}

	$session->close;

	foreach my $k ( sort keys %aifx ){								# lowest IPs first
		if(exists $main::int{$dv}{$aifx{$k}}){							# Avoid non existant IFs (e.g. idx=0 on  cisco2970 and 3750 with IOS 12.1)
			my @i		= split(/\./,$k);
			if (defined $i[13]){								# Some devs have incomplete IPs here!
				my $iaddr	= "$i[10].$i[11].$i[12].$i[13]";
				$main::net{$dv}{$iaddr}{ifn} = $main::int{$dv}{$aifx{$k}}{ina};
				$main::net{$dv}{$iaddr}{msk} = $ainm{"$ianmO.$iaddr"};
				if( exists $vrfna{$main::int{$dv}{$aifx{$k}}{ina}} ){
					$main::net{$dv}{$iaddr}{vrf} = $vrfna{$main::int{$dv}{$aifx{$k}}{ina}};
					$main::net{$dv}{$iaddr}{vrs} = $vrfst{$main::int{$dv}{$aifx{$k}}{ina}};
				}
				print "\n IP:$main::net{$dv}{$iaddr}{ifn}\t$iaddr/$main::net{$dv}{$iaddr}{msk}" if $main::opt{v};
				if($iaddr !~ /^127.0.0|^0/){
					if(exists $misc::ifip{$iaddr}){
						if($misc::notify =~ /a/){
							if( ! &db::Insert('messages','level,time,source,info',"\"150\",\"$main::now\",\"$dv\",\"IP $iaddr also detected on $misc::ifip{$iaddr}!\"") ){
								die "DB error messages!\n";
							}
						}
					}else{
						if ($main::int{$dv}{$aifx{$k}}{typ} == 24){		# 1.priority, use highest loopback IP
							if($ippri != 1){
								$newip = $iaddr;
								$ippri = 1;
							}elsif(&misc::Ip2Dec($iaddr) > &misc::Ip2Dec($newip)){
								$newip = $iaddr;
							}
						}elsif ($main::int{$dv}{$aifx{$k}}{typ} == 53 and $ippri > 3){		# 2.priority, use virtual IF(prefer existing IP)
							if($iaddr eq $main::dev{$dv}{ip}){
								$ippri = 3;
							}else{
								$ippri = 4;
							}
							$newip = $iaddr;
						}elsif ($main::int{$dv}{$aifx{$k}}{typ} =~ /^[67]$/ and $ippri > 5){	# 3.priority, use ethernet IF (prefer existing IP)
							if($iaddr eq $main::dev{$dv}{ip}){
								$ippri = 5;
							}else{
								$ippri = 6;
							}
							$newip = $iaddr;
						}
					}
					$misc::ifip{$iaddr} .= "$dv ";
				}
				$nia++;
			}
		}
	}
	if ($ippri < 10 and !$main::opt{I}){
		$main::dev{$dv}{oi} = $main::dev{$dv}{ip};
		$main::dev{$dv}{ip} = &misc::MapIp($newip);
		print "\n New IP:$main::dev{$dv}{ip} (Priority $ippri)" if $main::opt{v};
	}
	print "j$nia" if !$main::opt{v};
	print ($warn?" ":"\t");
}

#===================================================================
# Converts CDP capabilities to sys services alike format
# Parameters:	CDP services string
# Global:	-
# Return:	SNMP services decimal
#===================================================================
sub CDPCap2Sv {

	my $srv = 0;
	my $sv  = hex(unpack("C",substr($_[0],length($_[0])-1,1)));
	if ($sv & 1)		{$srv =   4}
	if ($sv & (8|4|2))	{$srv +=  2}
	if ($sv & 16)		{$srv += 64}
	if ($sv & 64)		{$srv +=  1}
	return $srv;
}

#===================================================================
# Use discovery protocol to find neighbours
# Parameters:	device name, discovery protocol id (usually name), discovery protocol
# Global:	main::dev, main::int, main::link, misc::portprop, (misc::doip if opt{c})
# Return:	-
#===================================================================
sub DisProtocol {

	use Socket;

	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";
	my $ad		= 0;
	my $dn		= 0;
	my $bd		= 0;

	my %neb  	= ();
	my %lpoe  	= ();

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
						-community => $main::dev{$_[0]}{cm},
						-timeout   => $misc::timeout * 2,			# some have longer (Vasily)
						-version   => $main::dev{$_[0]}{sp},
						-translate => [-octetstring => 0x0],
						-maxmsgsize=> 4095
						);

	if ($_[2] eq 'LLDP'){
		$r = $session->get_table('1.0.8802.1.1.2.1.5.4795.1.2.11.1.1');				# Get LLDP PoE
		$err = $session->error;
		if ($err){
			print "Qw";
			print "$err\n" if $main::opt{d};
		}else{
			while( my($key, $val) = each(%{$r}) ) {
				my @k = split (/\./,$key);
				$lpoe{$k[14]} = $val * 100;
			}
		}
		$r = $session->get_table('1.0.8802.1.1.2.1.4.2.1.3');					# Get LLDP IPs
		$err = $session->error;
		if ($err){
			print "Qa";
			print "$err\n" if $main::opt{d};
		}else{
			while( my($key, $val) = each(%{$r}) ) {
				my @k = split (/\./,$key);
				$neb{$k[12]}{$k[13]}{'ip'} = (defined $k[19])?&misc::MapIp("$k[16].$k[17].$k[18].$k[19]"):'';
			}
		}
		$r = $session->get_table('1.0.8802.1.1.2.1.4.1.1');
		while( my($key, $val) = each(%{$r}) ) {
			my @k = split (/\./,$key);
			if($k[10] == 7){								# lldpRemPortId
				if(${$r}{"1.0.8802.1.1.2.1.4.1.1.6.0.$k[12].$k[13]"} eq 7 and $val =~ /^[0-9]+$/){	# Prefer Descr if subtype is local and a #
					$neb{$k[12]}{$k[13]}{'if'} = &misc::Shif(${$r}{"1.0.8802.1.1.2.1.4.1.1.8.0.$k[12].$k[13]"});
				}elsif(${$r}{"1.0.8802.1.1.2.1.4.1.1.6.0.$k[12].$k[13]"} eq 3){		# if subtype is MAC address
					$neb{$k[12]}{$k[13]}{'if'} = unpack("H16",$val);
				}else{
					$neb{$k[12]}{$k[13]}{'if'} = &misc::Shif($val);
				}
				$neb{$k[12]}{$k[13]}{'pw'} = $lpoe{$k[12]};
			}elsif($k[10] == 9 and $val){							# lldpRemSysName
				$neb{$k[12]}{$k[13]}{'na'} = &misc::Strip($val);
			}elsif($k[10] == 5 and !$neb{$k[12]}{$k[13]}{'na'}){				# lldpRemChassisId if no name yet
#lldpRemChassisIdSubtype: Component(1),interfaceAlias(2),portComponent(3),macAddress(4),networkAddress(5),interfaceName(6),local(7)
#lldpRemPortIdSubtype: interfaceAlias(1), portComponent(2),macAddress(3),networkAddress(4),interfaceName(5),agentCircuitId(6),local(7)
				if(${$r}{"1.0.8802.1.1.2.1.4.1.1.4.0.$k[12].$k[13]"} eq 5){		# if subtype is IP address
					$neb{$k[12]}{$k[13]}{'ip'} = join('.',unpack('C*',substr($val,1) ));
					$neb{$k[12]}{$k[13]}{'na'} = gethostbyaddr(inet_aton($neb{$k[12]}{$k[13]}{'ip'}), AF_INET) or $neb{$k[12]}{$k[13]}{'na'} = join('-',unpack('C*',substr($val,1) ));
				}elsif(${$r}{"1.0.8802.1.1.2.1.4.1.1.4.0.$k[12].$k[13]"} eq 4){		# if subtype is MAC address
					$neb{$k[12]}{$k[13]}{'na'} = unpack("H16",$val);
				}else{
					$neb{$k[12]}{$k[13]}{'na'} = &misc::Strip($val);
				}
			}elsif($k[10] == 10){
				$neb{$k[12]}{$k[13]}{'de'} = &misc::Strip($val);
				$neb{$k[12]}{$k[13]}{'ty'} = $neb{$k[12]}{$k[13]}{'de'};		# No Type with LLDP :-(
			}elsif($k[10] == 11){
				$neb{$k[12]}{$k[13]}{'sv'} = unpack("C",$val);				# Not well supported!
			}
		}
	}elsif ($_[2] eq 'CDP'){
		$r = $session->get_table('1.3.6.1.4.1.9.9.23.1.2.1.1');
		while( my($key, $val) = each(%{$r}) ) {
			my @k = split (/\./,$key);
			if($k[13] == 4){
				if($val){
					$neb{$k[14]}{$k[15]}{'ip'} = &misc::MapIp(unpack("C",substr($val,0,1)).".".
										unpack("C",substr($val,1,1)).".".
										unpack("C",substr($val,2,1)).".".
										unpack("C",substr($val,3,1)) );
				}else{$neb{$k[14]}{$k[15]}{'ip'} = ''}
			}elsif($k[13] == 5){
				$neb{$k[14]}{$k[15]}{'de'} = &misc::Strip($val);
			}elsif($k[13] == 6){
				my $nebid = &misc::Strip($val);
				$neb{$k[14]}{$k[15]}{'id'} = $nebid;
				$neb{$k[14]}{$k[15]}{'na'} = $nebid;			
				$neb{$k[14]}{$k[15]}{'na'} =~ s/(.*?)\((.*?)\)/$2/;			# Extract from CatOS
				$neb{$k[14]}{$k[15]}{'na'} =~ s/(\xff){1,}/BadCDP-$k[15]/;		# Fixes some phone weirdness
			}elsif($k[13] == 7){
				$neb{$k[14]}{$k[15]}{'if'} = &misc::Shif($val);
			}elsif($k[13] == 8){
				$neb{$k[14]}{$k[15]}{'ty'} = &misc::Strip($val);
			}elsif($k[13] == 9){
				$neb{$k[14]}{$k[15]}{'sv'} = &CDPCap2Sv($val);
			}elsif($k[13] == 10){
				$neb{$k[14]}{$k[15]}{'vd'} = &misc::Strip($val);
			}elsif($k[13] == 12){
				if($val == 2){
					$neb{$k[14]}{$k[15]}{'dx'} = "HD";
				}elsif($val == 3){
					$neb{$k[14]}{$k[15]}{'dx'} = "FD";
				}
			}elsif($k[13] == 14){
				$neb{$k[14]}{$k[15]}{'vl'} = &misc::Strip($val);
			}elsif($k[13] == 15){
				$neb{$k[14]}{$k[15]}{'pw'} = &misc::Strip( ($val > 4000000000 or !$val)?0:$val );
			}
		}
	}elsif ($_[2] eq 'FDP'){
		$r = $session->get_table("1.3.6.1.4.1.1991.1.1.3.20.1.2");
		print "TBD!";
	}
	$err = $session->error;
	$session->close;

	if ($err and $err !~ 'OBJECT IDENTIFIER length equal to zero'){					# Give up on error
		print "Qp";
		print "=$err\n" if $main::opt{d};
		return 1;
	}
	foreach my $i ( keys %neb ){
		my $lif = $main::int{$_[0]}{$i}{ina};							# Assign interfacename.
		foreach my $n ( keys %{$neb{$i}} ){
			$neb{$i}{$n}{'na'} =~ s/^(.*?)\.(.*)/$1/ if !$main::opt{F};			# Strip domain
			print " $_[2]: $lif> $neb{$i}{$n}{'na'},$neb{$i}{$n}{'if'} IP=$neb{$i}{$n}{'ip'} TYP:$neb{$i}{$n}{'ty'} $neb{$i}{$n}{'de'} SV:$neb{$i}{$n}{'sv'} P:$neb{$i}{$n}{'pw'}mW on $neb{$i}{$n}{'if'}-$neb{$i}{$n}{'dx'} Vl$neb{$i}{$n}{'vl'}\n" if $main::opt{v};
			$main::int{$_[0]}{$i}{com} .= " $_[2]:$neb{$i}{$n}{'na'},$neb{$i}{$n}{'if'}";
			if ($_[1] eq $neb{$i}{$n}{'id'}){						# Seeing myself?
				$main::int{$_[0]}{$i}{com} .= " Loop!";
				$misc::portprop{$_[0]}{$lif}{upl} = 1;
				if($misc::notify =~ /d/){
					if( ! &db::Insert('messages','level,time,source,info',"\"150\",\"$main::now\",\"$_[0]\",\"$_[2]:Loop on $lif!\"") ){
						die "DB error messages!\n";
					}
				}
				print "Ql";
			}elsif($neb{$i}{$n}{'ip'} =~ /^$|^0\.0\.0\.0$|^127\.0\.0/){			# Not uplink in portprop, so nodes stay here...
				if($misc::notify =~ /d/){
					if( ! &db::Insert('messages','level,time,source,info',"\"100\",\"$main::now\",\"$_[0]\",\"$_[2]:$neb{$i}{$n}{'na'} IP=$neb{$i}{$n}{'ip'} on $neb{$i}{$n}{'if'}\"") ){
						die "DB error messages!\n";
					}
				}
				print "Qo";
			}else{
				if(!exists $main::link{$_[0]}{$lif}){					# No DP link if static exists
					$main::link{$_[0]}{$lif}{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{bw} = $main::int{$_[0]}{$i}{spd};
					$main::link{$_[0]}{$lif}{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{ty} = $_[2];
					$main::link{$_[0]}{$lif}{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{pr} = $neb{$i}{$n}{'pw'};
					$main::link{$_[0]}{$lif}{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{du} = $neb{$i}{$n}{'dx'};
					$main::link{$_[0]}{$lif}{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{vl} = $neb{$i}{$n}{'vl'};
					$main::int{$_[0]}{$i}{lty} = $_[2];
				}
				if("$neb{$i}{$n}{'de'}$neb{$i}{$n}{'ty'}" =~ /$misc::nosnmpdev/ or $neb{$i}{$n}{'na'} =~ /^AVT/){# Quick hack to check Avaya Phones
					print " noSNMP: Matching /$misc::nosnmpdev/\n" if $main::opt{v};
					$main::link{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{$_[0]}{$lif}{bw} = $main::int{$_[0]}{$i}{spd};
					$main::link{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{$_[0]}{$lif}{ty} = $_[2];
					$main::link{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{$_[0]}{$lif}{du} = $main::int{$_[0]}{$i}{dpx};
					$main::link{$neb{$i}{$n}{'na'}}{$neb{$i}{$n}{'if'}}{$_[0]}{$lif}{vl} = $main::int{$_[0]}{$i}{vln};

					$main::dev{$neb{$i}{$n}{'na'}}{ip} = $neb{$i}{$n}{'ip'};
					$main::dev{$neb{$i}{$n}{'na'}}{sn} = "-";
					$main::dev{$neb{$i}{$n}{'na'}}{bi} = "-";
					$main::dev{$neb{$i}{$n}{'na'}}{de} = $neb{$i}{$n}{'de'};
					$main::dev{$neb{$i}{$n}{'na'}}{sv} = $neb{$i}{$n}{'sv'};
					$main::dev{$neb{$i}{$n}{'na'}}{ty} = $neb{$i}{$n}{'ty'};
					$main::dev{$neb{$i}{$n}{'na'}}{os} = "-";
					$main::dev{$neb{$i}{$n}{'na'}}{lo} = $main::dev{$_[0]}{lo};
					$main::dev{$neb{$i}{$n}{'na'}}{co} = $main::dev{$_[0]}{co};
					$main::dev{$neb{$i}{$n}{'na'}}{vd} = $neb{$i}{$n}{'vd'};
					$main::dev{$neb{$i}{$n}{'na'}}{vm} = 0;
					if (!$main::dev{$neb{$i}{$n}{'na'}}{fs}){$main::dev{$neb{$i}{$n}{'na'}}{fs} = $main::now}
					$main::dev{$neb{$i}{$n}{'na'}}{ls} = $main::now;
					push (@misc::doneip,$neb{$i}{$n}{'ip'});
					if($neb{$i}{$n}{'ty'} =~ /Cisco IP Phone\s*(.*)$/){
						if($web::lwpok){
							&web::CiscoPhone($neb{$i}{$n}{'na'});
						}
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Cisco $1";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "phbn";
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 66;
						$misc::portprop{$_[0]}{$lif}{pho} = 1;
					}elsif($neb{$i}{$n}{'ty'} =~ /ATA/){
						if($web::lwpok){
							&web::CiscoAta($neb{$i}{$n}{'na'});
						}
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Cisco ATA Box";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "atbn";
						$misc::portprop{$_[0]}{$lif}{pho} = 1;
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 66;
					}elsif($neb{$i}{$n}{'ty'} =~ /Linksys IP Phone\s*(.*)$/){
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Linksys $1";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "phbn";
						$misc::portprop{$_[0]}{$lif}{pho} = 1;
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 66;
					}elsif($neb{$i}{$n}{'ty'} =~ /(MAP-.*)/){
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "HP $1";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "wagn";
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 2;
						$misc::portprop{$_[0]}{$lif}{wln} = 1;
					}elsif($neb{$i}{$n}{'ty'} =~ /AIR-BR/){
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Cisco Wlan Bridge";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "wbbn";
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 2;
						$misc::portprop{$_[0]}{$lif}{wln} = 1;
					}elsif($neb{$i}{$n}{'ty'} =~ /AIR-/){
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Cisco Wlan AP";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "wabn";
						$misc::portprop{$_[0]}{$lif}{wln} = 1;
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 2;
					}elsif($neb{$i}{$n}{'ty'} =~ /Nortel IP Telephone/){
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Nortel IP Phone";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "phon";
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 66;
						$misc::portprop{$_[0]}{$lif}{pho} = 1;
					}elsif($neb{$i}{$n}{'na'} =~ /^AVT/ and $neb{$i}{$n}{'sv'} eq 36){
						$main::dev{$neb{$i}{$n}{'na'}}{ty} = "Avaya IP Phone";
						$main::dev{$neb{$i}{$n}{'na'}}{ic} = "phan";
						$main::dev{$neb{$i}{$n}{'na'}}{sv} = 66;
						$misc::portprop{$_[0]}{$lif}{pho} = 1;
					}
					if (!$main::opt{t}){
						&db::WriteDev($neb{$i}{$n}{'na'});
						&db::WriteLink($neb{$i}{$n}{'na'});
					}
					push (@misc::doneip,$neb{$i}{$n}{'ip'});
				}else{
					$misc::portprop{$_[0]}{$lif}{upl} = 1;
					if(grep /^\Q$neb{$i}{$n}{'na'}\E$/,(@misc::doneid,@misc::failid,@misc::todo) or
					   grep /^\Q$neb{$i}{$n}{'ip'}\E$/,(@misc::doneip,@misc::failip) ){	# Don't add if done or already queued... (The \Q \E is to prevent interpreting the CDPid as a regexp)
						$dn++;
						print "DONE" if $main::opt{v};
					}elsif(defined $misc::border and $neb{$i}{$n}{'id'} =~ /$misc::border/){# or matching the border
						$bd++;
						print "BORDER" if $main::opt{v};
					}elsif($main::opt{p}){							# Add if discovery protocol ok
						push (@misc::todo,"$neb{$i}{$n}{'na'}");
						$misc::doip{$neb{$i}{$n}{'na'}} = $neb{$i}{$n}{'ip'};
						$ad++;
						print "ADDED" if $main::opt{v};
					}
				}
			}
		}
	}
	if ($_[2] and !$main::opt{v}){
		print "p$ad/$dn ";
		print "b$bd " if $bd;
	}
}

#===================================================================
# Get routing table information and queue next hop IPs
# Parameters:	device name
# Global:	main::dev, main::int, main::link, misc::portprop, misc::doip
# Return:	-
#===================================================================
sub Routes {
	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";

	my $warn	= 0;
	my $ad		= 0;
	my $dn		= 0;
	my $bd		= 0;

	my $cinhO	= "1.3.6.1.2.1.4.24.4.1.4";
	my $rtnhO	= "1.3.6.1.2.1.4.21.1.7";

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
						-community => $main::dev{$_[0]}{cm},
						-timeout   => $misc::timeout,
						-version   => $main::dev{$_[0]}{sp},
						-translate => [-octetstring => 0x0]
						);

	$r   = $session->get_table($cinhO);
	$err = $session->error;
	if ($err){
		print " CIDR:$err\n" if $main::opt{d};
		$r   = $session->get_table($rtnhO);							# Fallback to RFC1213
		$err = $session->error;
		if ($err){
			$warn++;
			print "Qr";
			print " RFC1213:$err\n" if $main::opt{d};
			$session->close;
			return 1;
		}
	}
	$session->close;

	while ( (my $key,my $val) =  each(%{$r}) ){
		my $nh = &misc::MapIp($val);								# Map IP if configured...
		print "\n RTE:$nh " if $main::opt{v};
		if (!exists $main::net{$_[0]}{$nh} and $nh ne "0.0.0.0"){				# IP is non-local or 0?
			if(grep /^\Q$nh\E$/,(@misc::doneid,@misc::doneip,@misc::failip,@misc::todo) ){	# Then queue if not done...
				$dn++;
				print " already done" if $main::opt{v};
			}elsif(defined $misc::border and $nh =~ /$misc::border/){			# ...or matching the border.
				$bd++;
				print " matches border" if $main::opt{v};
			}else{
				push (@misc::todo,"$nh");
				$misc::doip{$nh} = $nh;
				print " added" if $main::opt{v};
				$ad++;
			}
		}else{
			print " is local" if $main::opt{v};
		}
	}
	if (!$main::opt{v}){
		print " r$ad/$dn" if $main::opt{d};
		print "b$bd" if $bd;
	}
	print ($warn?" ":"\t");
}

#===================================================================
# Get ARP tables from Layer 3 device and queue entries for OUI discovery, if desired.
# Parameters:	device name
# Global:	misc::arp, misc::arpn, misc::arpc, misc::rarp, misc::portprop, misc::portnew, (misc::doip if opt{o})
# Return:	-
#===================================================================

sub Arp {

	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";

	my %at		= ();
	my %ntmtab	= ();
	my %myarpc	= ();

	my $warn	= 0;
	my $narp	= 0;
	my $ad		= 0;
	my $dn		= 0;
	my $bd		= 0;

	my $NmifO	= "1.3.6.1.2.1.4.22.1.2";

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
						-community => $main::dev{$_[0]}{cm},
						-timeout   => $misc::timeout,
						-version   => $main::dev{$_[0]}{sp},
						-translate => [-octetstring => 0x0]
						);

	$r   = $session->get_table($NmifO);
	$err = $session->error;
	$session->close;
	if ($err){print "Aa";print "$err\n" if $main::opt{d}}else{%at  = %{$r}}
	foreach my $k ( keys %at ){
		if ($k !~ /127.0.0/ and defined substr($at{$k},5,1) ){				# complete MAC and not loopback?
			my @i   = split(/\./,$k);
			my $mc   = unpack("H2",substr($at{$k},0,1)).unpack("H2",substr($at{$k},1,1)).unpack("H2",substr($at{$k},2,1)).unpack("H2",substr($at{$k},3,1)).unpack("H2",substr($at{$k},4,1)).unpack("H2",substr($at{$k},5,1));
			if ($mc !~/^f+$/){							# Not  broadcast?
				$misc::arp{$mc} = "$i[11].$i[12].$i[13].$i[14]";
				$misc::rarp{"$i[11].$i[12].$i[13].$i[14]"} = $mc;		# will be needed to identify OUI uplinks;
				$myarpc{$mc}++;							# Find multiple MACs
				my $po = "";
				if(exists $main::int{$_[0]}{$i[10]} ){
					$po = $main::int{$_[0]}{$i[10]}{ina};
					$misc::portprop{$_[0]}{$po}{rtr} = 1;
					$misc::portprop{$_[0]}{$po}{pop}++;
					$misc::portnew{$mc}{$_[0]}{po} = $po;
					$misc::portnew{$mc}{$_[0]}{vl} = ($po =~ /^Vl(\d+)$/) ? $1 : 0;
					print "\n ARP:$mc=$misc::arp{$mc} on $po" if $main::opt{v};
				}else{								# This can happen on loadbalancers, when IF is in AT table
					print "An";
					$warn++;
					$misc::portprop{$_[0]}{'-'}{rtr} = 1;
					$misc::portprop{$_[0]}{'-'}{chn} = 1;			# No IF, treat as channel to avoid MAC population
					$misc::portprop{$_[0]}{'-'}{pop}++;
					$misc::portnew{$mc}{$_[0]}{po} = '-';
				}
				$narp++;
				if ($main::opt{o}){
					my $oui = &misc::GetOui($mc);
					if($oui =~ /$misc::ouidev/i){
						if (grep /\Q$mc\E/,(@misc::doneid,@misc::failid,@misc::todo) ){# Don't queue if done or queued.
							$dn++;
						}elsif (defined $misc::border and $mc =~ /$misc::border/){# ...or matching the border...
							$bd++;
						}elsif ($misc::arp{$mc} eq '0.0.0.0'){		# ...or no IP found
							if($misc::notify =~ /d/){
								if( ! &db::Insert('messages','level,time,source,info',"\"100\",\"$main::now\",\"$mc\",\"OUI device ($oui) IP=0.0.0.0\"") ){
									die "DB error messages!\n";
								}
							}
						}else{
							push (@misc::todo,"$mc");
							$misc::doip{$mc} = &misc::MapIp($misc::arp{$mc});
							print "\n OUI:$po\t$oui ($misc::arp{$mc}) " if $main::opt{v};
							$ad++;
						}
					}
				}
			}
		}
	}

	foreach my $mc ( keys %myarpc ){
		$misc::arpc{$mc} = $myarpc{$mc};
		if( $myarpc{$mc} > $misc::arppoison and $misc::notify =~ /n/i){	# Many ARPs, poisoning?
			if( ! &db::Insert('messages','level,time,source,info',"\"150\",\"$main::now\",\"$mc\",\"$_[0] has $myarpc{$mc} ARP entries for $misc::arp{$mc}\"") ){
				die "DB error messages!\n";
			}
			print &mon::SendMail("ARP Poisoning Alert","$_[0] has $myarpc{$mc} ARP entries for $misc::arp{$mc} ($mc)!") if($misc::notify =~ /N/);
		}
	}

	if (!$main::opt{v}){
		print " a$narp";
		print " o$ad/$dn" if $main::opt{d};
		print "b$bd" if $bd;
	}
	print ($warn?" ":"\t");
}

#===================================================================
# Converts MSM IF type to IEEE types
# Parameters:	Colubris type
# Global:	-
# Return:	IEEE type
#===================================================================
sub MSM2I {

	if ($_[0] == 2){
		return 6;
	}elsif ($_[0] == 3){
		return 53;
	}elsif ($_[0] == 4){
		return 209;
	}elsif ($_[0] == 5){
		return 71;
	}else{
		return $_[0];
	}
}

#===================================================================
# Get MAC address table from a device with optional community indexing.
# Parameters:	device name
# Global:	misc::portprop, misc::portnew
# Return:	-
#===================================================================

sub BridgeFwd {

	my $session	= "";
	my $error	= "";
	my $r		= "";
	my $err		= "";

	my $nspo	= 0;
	my @vlans	= ();

	my $fwdxO  = '1.3.6.1.2.1.17.1.4.1.2';
	my $fwdpO  = '1.3.6.1.2.1.17.4.3.1.2';
	my $fwdsO  = '1.3.6.1.2.1.17.5.1.1.1';

	if($misc::sysobj{$main::dev{$_[0]}{so}}{bf} eq "MSM"){
		my $distO  = '1.3.6.1.4.1.8744.5.24.1.1.1.1';					# 2=name,3=type,4=vlid,5,=ip,6=mask,7=mac,8=state
		my $ditrO  = '1.3.6.1.4.1.8744.5.24.1.2.1.1';					# 1=ioct,3=ierr,4=ooct,6=oerr
		my $iwixO  = '1.3.6.1.4.1.8744.5.25.1.2.1.1.2';					# if to wireless-if index
		my $diclO  = '1.3.6.1.4.1.8744.5.25.1.7.1.1';					# 2=mac,3=vsc,17=ip
		my %iwx    = ();

		($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
							-community => $main::dev{$_[0]}{cm},
							-timeout   => $misc::timeout * 2,
							-version   => $main::dev{$_[0]}{sp},
							-translate => [-octetstring => 0x0]
							);

		$r   = $session->get_table($distO);
		$err = $session->error;
		if ($err){
			print "Fwi";
			print "$err\n" if $main::opt{d};
			$session->close;
			return;
		}else{
			my %ifnet	= ();
			while( my($key, $val) = each(%{$r}) ) {					# Get general IF values first
				my @k = split (/\./,$key);
				$i = sprintf("%d%02d",$k[14],$k[15]);				# Map AP,Ant to new IF index
				if($k[13] == 2){
					$main::int{$_[0]}{$i}{ina} = "$main::mod{$_[0]}{$k[14]}{fw}/".&misc::Shif($val);
				}elsif($k[13] == 3){
					$main::int{$_[0]}{$i}{typ} = &MSM2I($val);
				}elsif($k[13] == 4){
					$main::int{$_[0]}{$i}{vln} = $val;
				}elsif($k[13] == 5 and $val ne "0.0.0.0"){
					$ifnet{$i}{ip} = $val;					# Table is not ordered, so collect IPs...
				}elsif($k[13] == 6 and $val ne "0.0.0.0"){
					$ifnet{$i}{mk} = $val;					# ...and then the masks...
				}elsif($k[13] == 7){
					my $mc = unpack('H12', $val);
					$main::int{$_[0]}{$i}{mac} = $mc;
					$misc::ifmac{$mc}++;
				}elsif($k[13] == 8){
					$main::int{$_[0]}{$i}{sta} = $val;
				}
			}
			foreach my $i (keys %ifnet){						# ...then map to the networks!
				$main::net{$_[0]}{$ifnet{$i}{ip}}{ifn} = $main::int{$_[0]}{$i}{ina};
				$main::net{$_[0]}{$ifnet{$i}{ip}}{msk} = $ifnet{$i}{mk};
			}
			$r    = $session->get_table($ditrO);
			$err = $session->error;
			if ($err){print "Fwc";print "$err\n" if $main::opt{d}}else{
				while( my($key, $val) = each(%{$r}) ) {				# ...then some traffic counters
					my @k = split (/\./,$key);
					$i = sprintf("%d%02d",$k[14],$k[15]);			# Map AP,Ant to new IF index
					if($k[13] == 1){
						$main::int{$_[0]}{$i}{ioc} = $val;
					}elsif($k[13] == 3){
						$main::int{$_[0]}{$i}{ier} = $val;
					}elsif($k[13] == 4){
						$main::int{$_[0]}{$i}{ooc} = $val;
					}elsif($k[13] == 6){
						$main::int{$_[0]}{$i}{oer} = $val;
					}
				}
			}
			$r    = $session->get_table($iwixO);
			$err = $session->error;
			if ($err){print "Fwx";print "$err\n" if $main::opt{d}}else{
				while( my($key, $val) = each(%{$r}) ) {				# Get IF index to Wlan-IF mapping ...
					my @k = split (/\./,$key);
					$iwx{$k[15]} = $val;
				}
			}
			my %cltmc	= ();
			my %cltvs	= ();
			my %cltip	= ();
			$r    = $session->get_table("$diclO.2");				# Get client MACs...
			$err = $session->error;
			if ($err){print "Fwm";print "$err\n" if $main::opt{d}}else{
				while( my($key, $val) = each(%{$r}) ) {
					my @k = split (/\./,$key);
					$cltmc{"$k[14].$k[15].$k[16]"} = unpack('H12', $val);
				}
			}
			$r    = $session->get_table("$diclO.3");				# Get client VSCs...
			$err = $session->error;
			if ($err){print "Fwv";print "$err\n" if $main::opt{d}}else{
				while( my($key, $val) = each(%{$r}) ) {
					my @k = split (/\./,$key);
					$cltvs{"$k[14].$k[15].$k[16]"} = $val;
				}
			}
			$r    = $session->get_table("$diclO.17");				# Get client IPs...
			$err = $session->error;
			if ($err){print "Fwa";print "$err\n" if $main::opt{d}}else{
				while( my($key, $val) = each(%{$r}) ) {
					my @k = split (/\./,$key);
					$cltip{"$k[14].$k[15].$k[16]"} = $val;
				}
			}
			$session->close;
			foreach my $key (keys %cltmc){						# ...and put'em all together
				my @k = split (/\./,$key);
				$i = sprintf("%d%02d",$k[0],$iwx{$k[1]});			# Map Wlan-IF to new IF index
				$misc::portnew{$cltmc{$key}}{$_[0]}{vl} = $cltvs{$key};
				$misc::portprop{$_[0]}{$main::int{$_[0]}{$i}{ina}}{pop}++;
				$misc::portnew{$cltmc{$key}}{$_[0]}{po} = $main::int{$_[0]}{$i}{ina};
				$misc::arp{$cltmc{$key}} = $cltip{$key} if $cltip{$key} != "0.0.0.0";
				print "\n FWM:$cltmc{$key},$cltip{$key} on $main::int{$_[0]}{$i}{ina} Vl$cltvs{$key} " if $main::opt{v};
				$nspo++;
			}
		}
		print "f$nspo" if !$main::opt{v};
		return;
	}elsif($misc::sysobj{$main::dev{$_[0]}{so}}{bf} eq "VLX"){
		@vlans = keys %{$main::vlan{$_[0]}};
	}else{
		$vlans[0] = "";
	}

	foreach my $vl (@vlans){
		if ($vl !~ /$misc::ignoredvlans/){
			my %fwdix = ();
			my %fwdpo = ();
			my %Fdpst = ();

			($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
								-community => $main::dev{$_[0]}{cm}.($vl ne ''?"\@$vl":""),
								-timeout   => $misc::timeout,
								-version   => $main::dev{$_[0]}{sp},
								-translate => [-octetstring => 0x0]
								);

			$r = $session->get_table($fwdxO);
			$err = $session->error;
			if ($err){print "Fi$vl ";print "$err\n" if $main::opt{d}}else{%fwdix = %{$r} }
			$r = $session->get_table($fwdpO);
			$err = $session->error;
			if ($err){print "Fp$vl ";print "$err\n" if $main::opt{d}}else{%fwdpo = %{$r} }
			$session->close;

			foreach my $fpo ( keys %fwdpo ){
				my @dmac = split(/\./,$fpo);
				if (defined $dmac[16]){						# Some have incomplete MACs!
					my $mc   = sprintf "%02x%02x%02x%02x%02x%02x",$dmac[11],$dmac[12],$dmac[13],$dmac[14],$dmac[15],$dmac[16];
					my $ifx  = $fwdix{"$fwdxO.$fwdpo{$fpo}"};
					if (defined $ifx){
						my $po = "fwd$ifx";				# Fallback for weird switches...
						if (defined $main::int{$_[0]}{$ifx}){
							$po   = $main::int{$_[0]}{$ifx}{ina};	# ...otherwhise use real name.

							if ($po =~ /[0-9]-[0-9]|[0-9],[0-9]|^Po[0-9]|channel/){
								$misc::portprop{$_[0]}{$po}{chn} = 1;
							}
						}
						$misc::portprop{$_[0]}{$po}{pop}++;
						$misc::portnew{$mc}{$_[0]}{po} = $po;
						if($vl){
							$misc::portnew{$mc}{$_[0]}{vl} = $vl;
						}else{
							$misc::portnew{$mc}{$_[0]}{vl} = $misc::portprop{$_[0]}{$po}{vln};
						}
						print "\n FWS:$mc on $po Vl$misc::portnew{$mc}{$_[0]}{vl} " if $main::opt{v};
						$nspo++;
					}else{
						print "No Ifix:$mc\n" if $main::opt{v};		# happens for switch's own MAC
					}
				}
			}
		}
	}
	print "f$nspo" if !$main::opt{v};
}

#===================================================================
# Get module list according to .def file.
# Parameters:	device name
# Global:	main::mod
# Return:	-
#===================================================================

sub Modules {

	my $warn = 0;
	my $nmod = 0;

	my %mde = ();
	my %mcl = ();
	my %msl = ();
	my %mhw = ();
	my %msw = ();
	my %mfw = ();
	my %msn = ();
	my %mmo = ();

	my $so	= $main::dev{$_[0]}{so};

	($session, $error) = Net::SNMP->session(-hostname  => $main::dev{$_[0]}{ip},
						-community => $main::dev{$_[0]}{cm},
						-timeout   => $misc::timeout,
						-version   => $main::dev{$_[0]}{sp},
						-translate => [-octetstring => 0x0]
						);

	$r = $session->get_table($misc::sysobj{$so}{mt});						# Walk slot/supplyclass
	$err = $session->error;
	if ($err){
		print "Mt";
		print "$err\n" if $main::opt{d};
		$warn++;
	}else{
		%msl  = %{$r};
		if($misc::sysobj{$so}{md}){
			$r = $session->get_table($misc::sysobj{$so}{md});				# Walk module description
			$err = $session->error;
			if ($err){print "Mt";print "$err\n" if $main::opt{d};return 1;}else{%mde  = %{$r}}
		}
		if($misc::sysobj{$so}{mc}){
			$r = $session->get_table($misc::sysobj{$so}{mc});				# Walk module classes
			$err = $session->error;
			if ($err){print "Mc";print "$err\n" if $main::opt{d};$warn++}else{%mcl  = %{$r}}
		}
		if($misc::sysobj{$so}{mh}){
			$r = $session->get_table($misc::sysobj{$so}{mh});				# Walk module HW/supply capacity
			$err = $session->error;
			if ($err){print "Mh";print "$err\n" if $main::opt{d};$warn++}else{%mhw  = %{$r}}
		}
		if($misc::sysobj{$so}{ms}){
			$r = $session->get_table($misc::sysobj{$so}{ms});				# Walk module software version
			$err = $session->error;
			if ($err){print "Ms";print "$err\n" if $main::opt{d};$warn++}else{%msw  = %{$r}}
		}
		if($misc::sysobj{$so}{mf}){
			$r = $session->get_table($misc::sysobj{$so}{mf});				# Walk module FW/supply level
			$err = $session->error;
			if ($err){print "Mf";print "$err\n" if $main::opt{d};$warn++}else{%mfw  = %{$r}}
		}
		if($misc::sysobj{$so}{mn}){
			$r = $session->get_table($misc::sysobj{$so}{mn});				# Walk module serial number
			$err = $session->error;
			if ($err){print "Mn";print "$err\n" if $main::opt{d};$warn++}else{%msn  = %{$r}}
		}
		if($misc::sysobj{$so}{mm}){
			$r = $session->get_table($misc::sysobj{$so}{mm});				# Walk module model
			$err = $session->error;
			if ($err){print "Mm";print "$err\n" if $main::opt{d};$warn++}else{%mmo  = %{$r}}
		}
	}
	$session->close;

	foreach my $i ( keys %msl ){
		my $ismod = 0;
		my $s = $msl{$i};
		$i =~ s/$misc::sysobj{$so}{mt}\.//;
		if (exists $mcl{"$misc::sysobj{$so}{mc}.$i"}){
			if($mcl{"$misc::sysobj{$so}{mc}.$i"} =~ /$misc::sysobj{$so}{mv}/){
				if($mcl{"$misc::sysobj{$so}{mc}.$i"} == 10){				# Class 10 and a SN is most likely a transceiver
					$ismod =1 if $msn{"$misc::sysobj{$so}{mn}.$i"};
				}else{
					$ismod =1;
				}
			}
		}else{
			$ismod = 1;
		}
		if($ismod){
			if ($main::dev{$_[0]}{os} eq "Printer"){
				$main::mod{$_[0]}{$i}{sl} = &misc::Strip($s);
				$main::mod{$_[0]}{$i}{de} = ($mde{"$misc::sysobj{$so}{md}.$i"}?$mde{"$misc::sysobj{$so}{md}.$i"}:"-");
				$main::mod{$_[0]}{$i}{hw} = ($mfw{"$misc::sysobj{$so}{mf}.$i"}?int(100*$mhw{"$misc::sysobj{$so}{mh}.$i"} / $mfw{"$misc::sysobj{$so}{mf}.$i"}):"0");
				$main::mod{$_[0]}{$i}{mo} = "Printsupply";
			}elsif ($mmo{"$misc::sysobj{$so}{mm}.$i"} or $mde{"$misc::sysobj{$so}{md}.$i"}){
				$main::mod{$_[0]}{$i}{sl} = substr($s,0,63);
				$main::mod{$_[0]}{$i}{mo} = (($mmo{"$misc::sysobj{$so}{mm}.$i"} and $mmo{"$misc::sysobj{$so}{mm}.$i"} ne $msn{"$misc::sysobj{$so}{mn}.$i"})?&misc::Strip($mmo{"$misc::sysobj{$so}{mm}.$i"}):"-");	# Some transceivers report serial as model (rufer)
				$main::mod{$_[0]}{$i}{de} = ($mde{"$misc::sysobj{$so}{md}.$i"}?&misc::Strip($mde{"$misc::sysobj{$so}{md}.$i"}):"-");
				$main::mod{$_[0]}{$i}{hw} = ($mhw{"$misc::sysobj{$so}{mh}.$i"}?&misc::Strip($mhw{"$misc::sysobj{$so}{mh}.$i"}):"-");
				$main::mod{$_[0]}{$i}{fw} = ($mfw{"$misc::sysobj{$so}{mf}.$i"}?&misc::Strip($mfw{"$misc::sysobj{$so}{mf}.$i"}):"-");
				$main::mod{$_[0]}{$i}{sw} = ($msw{"$misc::sysobj{$so}{ms}.$i"}?&misc::Strip($msw{"$misc::sysobj{$so}{ms}.$i"}):"-");
				$main::mod{$_[0]}{$i}{sn} = ($msn{"$misc::sysobj{$so}{mn}.$i"}?&misc::Strip($msn{"$misc::sysobj{$so}{mn}.$i"}):"-");
				$main::mod{$_[0]}{$i}{st} = 0;
			}
			print "\n MOD:$i $s\t$main::mod{$_[0]}{$i}{mo}\t$main::mod{$_[0]}{$i}{de}"  if $main::opt{v};
			$nmod++;
		}
	}
	print "m$nmod" if !$main::opt{v};
	print ($warn?" ":"\t");
}

1;
