#============================================================================
#
# Program: libweb.pl
# Programmer: Remo Rickli, dcr
#
# -> LWP based Functions <-
#
# Fetches info from supported web only devices 
#============================================================================
package web;

use vars qw($lwpok);

eval 'use LWP::UserAgent;';
if ($@){
	$lwpok = 0;
	print "LWP not available\n" if $main::opt{d};
}else{
	$lwpok = 1;
	print "LWP loaded\n" if $main::opt{d};
}

#============================================================================
# Fetch info through webinterface of Cisco phones (by Remo Rickli & Kyle Kniffin)
# Parameters:	device name
# Global:	main::dev
# Return:	0 on success, 1 on failure
#============================================================================
sub CiscoPhone{
	my $ua = LWP::UserAgent->new;
	$ua->timeout($misc::timeout);
	my $response = $ua->get("http://$main::dev{$_[0]}{ip}");
	if ($response->is_success) {
		my $devhtml = $response->content;
		if ( $devhtml =~ m/<b>([0-9]{2,8})<\/b>/i ) {				# Find Extension on Cisco 79-20,40,60,11,10,06
			$main::dev{$_[0]}{co} = $1;
		}elsif( $devhtml =~ m/<td>([0-9]{2,8})\W+<tr>/i ){			# Find Extension on Cisco 7912
			$main::dev{$_[0]}{co} = $1;
		}
		if ( $devhtml =~ m/<b>(FCH\w+|INM\w+)<\/b>/i ) {			# Find Serial  on Cisco 79-20,40,60,11,10,06
			$main::dev{$_[0]}{sn} = $1;
		}elsif( $devhtml =~ m/<td>(FCH\w+|INM\w+)\W+<tr>/i ){			# Find Serial  on Cisco 7912
			$main::dev{$_[0]}{sn} = $1;
		}
		print " LWP:$_[0]-$main::dev{$_[0]}{co} SN:$main::dev{$_[0]}{sn}\n" if $main::opt{v};
	} else {
		print " LWP:" . $response->status_line if $main::opt{d};
		return 1;
	}
}

#============================================================================
# Fetch info through webinterface of Cisco ATA boxes (by Kyle Kniffin)
# Parameters:	device name
# Global:	main::dev
# Return:	0 on success, 1 on failure
#============================================================================
sub CiscoAta{

	my $ua = LWP::UserAgent->new;
	$ua->timeout($misc::timeout);

	my $response = $ua->get("http://$main::dev{$_[0]}{ip}/DeviceInfo");
	if ($response->is_success) {
		my $devhtml = $response->content;
		if ( $devhtml =~ m/<td>([0-9]{2,8})<\/td>/i ) {				# Find Extension on Cisco ATA 186
			$main::dev{$_[0]}{co} = $1;
		}
		if ( $devhtml =~ m/<td>(FCH\w+|INM\w+)<\/td>/i ) {			# Find Serial  on Cisco ATA 186
			$main::dev{$_[0]}{sn} = $1;
		}
		print "\n LWP:$_[0]-$main::dev{$_[0]}{co} SN:$main::dev{$_[0]}{sn}" if $main::opt{v};
	} else {
		print " LWP:" . $response->status_line if $main::opt{d};
		return 1;
	}
 }
	
1;
