#!/usr/bin/perl
#============================================================================
# Program: moni.pl
# Programmer: Remo Rickli
#
# DATE     COMMENT
# ------------------------------------------------------------------------
# 15/06/05	initial version
# 17/05/06	non-blocking uptime queries and optimized dependency resolution
# 17/01/07	updated path handling
# 04/06/07	improved stability
# 04/09/07	added test function
# 11/03/08	No dependency changed from none to - and fixed disappearing dependencies
# 28/11/09	Refer to NeDi Forum for changes...

# REDMINE power is in your hands

$VERSION = "1.0.5";

#============================================================================
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#============================================================================
# Visit http://nedi.sourceforge.net for more information.
#============================================================================

use strict;
use Getopt::Std;
use Net::SNMP qw(snmp_dispatcher ticks_to_time);
use Time::HiRes;

use vars qw($p $now $maxi);										# Global since access from functions is required
use vars qw(%opt %dev %usr %mon %depdevs %depdown %depcount);

getopts('Dt:vV',\%opt) || &HELP_MESSAGE;

$p = $0;
$p =~ s/(.*)\/(.*)/$1/;
if($0 eq $p){$p = "."};
require "$p/inc/libmisc.pl";										# Include required libraries
&misc::ReadConf();
require "$p/inc/libdb-" . lc($misc::backend) . ".pl" || die "Backend error ($misc::backend)!";
require "$p/inc/libmon.pl";

if ($opt{t}) {												# Creates incident and bails
	$now = time;
	my $msg = "moni.pl test message";
	if( ! &db::Insert('messages','level,time,source,info',"\"$opt{t}\",\"$now\",\"Test\",\"$msg\"") ){
		die "DB error messages!\n";
	}
	if( ! &db::Insert('incidents','level,device,deps,firstseen,lastseen,who,time,category,comment',"\"$opt{t}\",\"Test\",\"0\",\"$now\",\"".($now+$misc::pause)."\",\"\",\"0\",\"1\",\"\"") ){
		die "DB error incidents!\n";
	}
	my $nmail = 0;
	my $nsms = 0;
	if( $opt{t} >= $misc::msglevel ){
		$nsms  = &mon::SendSMS($msg);
		$nmail = &mon::SendMail("moni.pl Test",$msg);
	}
	die "1 Message, 1 Incident, $nsms SMS and $nmail mail created\n";
}elsif ($opt{D}) {											# Daemonize or...
	&misc::Daemonize;
}else{
	select(STDOUT); $| = 1;										# ...disable buffering.
}

#my $devup	= 0;
while(1) {
	$now = time;
	print "Scan:\t" . localtime($now) . "\n" if $opt{v};
#	if($now - $misc::rrdstep > $devup){								# refresh devices if older than a discovery...
#		$devup = $now;
		&db::ReadDev();
#	}
	&db::ReadMon();
	$maxi  = 100;											# maximum iterations to prevent potential loops upon resolving misconfigured dependencies
	%depdown  = ();											# dead dependencies are set to 1
	%depdevs  = ();											# list of dependant devices
	%depcount = ();											# count of dependants

	print "==================================================================================\n" if $opt{v};

	foreach my $d (keys %mon){
		if(defined $dev{$d}{cm}){								# Query only if device has a community set
			if( !($mon{$d}{dp} eq '-' or $mon{$d}{dp} eq 'none') ){				# Does it have a dependency configured?
				if(exists $dev{$mon{$d}{dp}}){						# And does it  exist in devices?
					push (@{$depdevs{$mon{$d}{dp}}},$d);				# Add it to dependency tree
				}else{
					my %set = ();
					$set{'depend'} = '-';
					if( ! &db::Update('monitoring',\%set,"device =\"$d\"") ){	# Remove dependency and notify
						die "DB error monitoring!\n";
					}
					if( ! &db::Insert('messages','level,time,source,info',"\"50\",\"$now\",\"$d\",\"Non existant dependency $mon{$d}{dp} removed.\"") ){
						die "DB error messages!\n";
					}
				}
			}
			my ($session, $error) = Net::SNMP->session(					# Get uptime
						-hostname    => $main::dev{$d}{ip},
						-community   => $main::dev{$d}{cm},
						-nonblocking => 0x1,   					# Create non-blocking objects
						-timeout     => $misc::timeout * 2,			# double to avoid false alarms (Vasily)
						-retries     => 1,
						-version     => $main::dev{$d}{sp},
						-translate   => [
						-timeticks => 0x0					# Turn off so sysUpTime is numeric
						]
			);
			if (!defined($session)) {
				printf("ERROR: %s.\n", $error);
				exit 1;
			}
			$mon{$d}{sta} = Time::HiRes::time;						# Start time to measure delay
			$session->get_request(	-varbindlist => ['1.3.6.1.2.1.1.3.0'],
						-callback    => [\&ProcessUpt, $d]
			);
		}
	}
	snmp_dispatcher();										# Shoots out the queries above
	print "----------------------------------------------------------------------------------\n" if $opt{v};
	print "Dependant Count:\n" if $opt{v};
	foreach my $d (keys %depdevs){									# Count dependencies to optimize processing
		$depcount{$d} = &CountDep($d,0);
		print " = $depcount{$d}\n" if $opt{v};
	}
	print "Optimized Processing:\n" if $opt{v};
	foreach my $d (sort { $depcount{$b} <=> $depcount{$a} } keys %mon){				# Now let's have a look at the answers
		print "$d\t" if $opt{v};
		if($mon{$d}{nu}){									# Did we get an uptime?
			print "Up " if $opt{v};
			if($depdown{$d}){								# impossible status -> notify
				print ", but parent is down! " if $opt{v};
				if( ! &db::Insert('messages','level,time,source,info',"\"50\",\"$now\",\"$d\",\"Up, although parent is down?\"") ){
					die "DB error messages!\n";
				}
				&MarkDep($d,0,0);
			}
			if($mon{$d}{st} >= $misc::thres){
				print "Recovered! " if $opt{v};
				if( ! &db::Insert('messages','level,time,source,info',"\"50\",\"$now\",\"$d\",\"Recovered (down since ".localtime($mon{$d}{lc})."\"") ){
					die "DB error messages!\n";
				}
				if( $mon{$d}{ml}){&mon::SendMail("Moni Recovery","Device $d Recovered!")}
				my %set = ();
				$set{'lastseen'} = $now;
				if( ! &db::Update('incidents',\%set,"device =\"$d\" AND lastseen=0") ){
					die "DB error incidents!\n";
				}
			}
			$mon{$d}{ok}++;
			my %set = ();
			$set{'status'}  = '0';
			$set{'lastchk'} = $now;
			$set{'uptime'}  = $mon{$d}{nu};
			$set{'ok'}      = $mon{$d}{ok};
			$set{'delay'}	= $mon{$d}{dly};
			if( ! &db::Update('monitoring',\%set,"device =\"$d\"") ){
				die "DB error monitoring!\n";
			}
			if($mon{$d}{ut} > $mon{$d}{nu} and $mon{$d}{ut} < 4294900000){			# ignore potential reboot, when close to the max 32bit value!
				print "Rebooted! " if $opt{v};
				if( ! &db::Insert('messages','level,time,source,info',"\"150\",\"$now\",\"$d\",\"Rebooted (was up for ". ticks_to_time($mon{$d}{ut}) .")!\"") ){
					die "DB error messages!\n";
				}
				if( $mon{$d}{ml}){&mon::SendMail("Moni Reboot Alert!","Device $d Rebooted!")}
			}
		}else{
			print "No answer, affects: " if $opt{v};
			&MarkDep($d,1,0);
			$mon{$d}{st}++;
			$mon{$d}{lt}++;
			my %set = ();
			$set{'status'}  = $mon{$d}{st};
			$set{'lost'}    = $mon{$d}{lt};
			#$set{'lastchk'} = $now;	No update on failure (Vasily)
			if( ! &db::Update('monitoring',\%set,"device =\"$d\"") ){
				die "DB error monitoring!\n";
			}
			if(!$depdown{$d} and $mon{$d}{st} == $misc::thres){
				print "For $mon{$d}{st} times! " if $opt{v};
				my $lvl = 200;
				my $cr  = "";
				if( $dev{$d}{lo} =~ /$misc::redbuild/){$lvl = 250;$cr = "Emergency"}
				my $downmsg = "Down!";
				if($depcount{$d}){
					$downmsg = "Down, affecting $depcount{$d} devices!!";
				}else{
					$depcount{$d} = 0;						# Prevents potential DB error
				}
				if( ! &db::Insert('messages','level,time,source,info',"\"$lvl\",\"$now\",\"$d\",\"$downmsg (no answer for $misc::thres times) \"") ){
					die "DB error messages!\n";
				}
				if( ! &db::Insert('incidents','level,device,deps,firstseen,lastseen,who,time,category,comment',"\"$lvl\",\"$d\",\"$depcount{$d}\",\"$now\",\"0\",\"\",\"0\",\"1\",\"\"") ){
					die "DB error incidents!\n";
				}
				if( $mon{$d}{ss}){&mon::SendSMS("$now: $d is down!")}
				if( $mon{$d}{ml}){&mon::SendMail("Moni $cr Alert!","Device $d is down!")}
			}

		}
		print "\n" if $opt{v};
	}
	print "\n==================================================================================\n" if $opt{v};
	sleep($misc::pause);
}

#===================================================================
# Callback function for non blocking SNMP uptime and calculating delay
# Parameters:	session, device name
# Global:		main::mon
# Return:		-
#===================================================================
sub ProcessUpt{

	my ($session, $d) = @_;
	my $errstat = $session->error_status;
	my $err = $session->error;

	print "$d\t" if $opt{v};
	if (!defined($session->var_bind_list)) {
		print "down $err($errstat), affects:" if $opt{v};
		&MarkDep($d,1,0);
		$mon{$d}{dly} = 0;
		$mon{$d}{nu} = 0;
	} else {
		$mon{$d}{nu} = $session->var_bind_list->{'1.3.6.1.2.1.1.3.0'};
		print "UP:".ticks_to_time($mon{$d}{nu}) if $opt{v};
		$mon{$d}{dly} = int(1000 * (Time::HiRes::time - $mon{$d}{sta}) );
		print " DLY:$mon{$d}{dly}ms\n" if $opt{v};
	}
}

#===================================================================
# Mark dependencies
# Parameters:	device name, up/down, iteration
# Global:		main::depdown
# Return:		-
#===================================================================
sub MarkDep {

	if($_[2] < $maxi and exists $depdevs{$_[0]} ){
		foreach my $d (@{$depdevs{$_[0]}}){
			print " $d" if $opt{v};
			$depdown{$d} = $_[1];
			&MarkDep($d,$_[1],$_[2]+1);
		}
	}
}

#===================================================================
# Recursively count dependants
# Parameters:	device name, iteration
# Global:		-
# Return:		# of dependees
#===================================================================
sub CountDep {
	if($_[1] < $maxi and exists $depdevs{$_[0]} ){
		my $c = scalar @{$depdevs{$_[0]}};
		print "$_[1]:$_[0]+$c " if $opt{v};
		foreach my $d (@{$depdevs{$_[0]}}){
			$c += &CountDep($d,$_[1]+1);
		}
		return $c;
	}else{
		return 0;
	}
}

#===================================================================
# Display some help
# Parameters:	-
# Global:		-
# Return:		-
#===================================================================
sub HELP_MESSAGE {
	print "\n";
	print "usage: moni.pl <Option(s)>\n\n";
	print "---------------------------------------------------------------------------\n";
	print "Options:\n";
	print "-t <lev>	test (creates mail (on msgxxx match), SMS, message and incident)\n";
	print "-v	verbose output\n";
	print "-D	daemonize moni.pl\n\n";
	print "Moni $main::VERSION (C) 2001-2009 Remo Rickli (and contributors)\n\n";
	die;
}
