#!/usr/bin/perl
#============================================================================
# Program: syslog.pl
# Programmer: Remo Rickli
#
# DATE     		COMMENT
# --------------------------------------------------------------------------
# 21/07/05	initial version
# 2/03/06	sanitized and performance optimized message handling
# 17/01/07	updated path handling
# 03/02/08 	limited msg size
# 28/03/08 	improved updating of devices
# 28/11/09	Refer to NeDi Forum for changes...

$VERSION = "1.0.5";

#============================================================================
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#============================================================================
# Visit http://nedi.sourceforge.net for more information.
#============================================================================

use strict;
use IO::Socket;
use Getopt::Std;

use vars qw($nediconf);
use vars qw($p %opt %dev %dip %usr);

getopts('Dvp:',\%opt)  || &HELP_MESSAGE;

$p = $0;
$p =~ s/(.*)\/(.*)/$1/;
if($0 eq $p){$p = "."};
require "$p/inc/libmisc.pl";										# Use the miscellaneous nedi library
&misc::ReadConf();
require "$p/inc/libmon.pl";										# Use the SNMP function library
require "$p/inc/libdb-" . lc($misc::backend) . ".pl" || die "Backend error ($misc::backend)!";

#Disable buffering.
select(STDOUT); $| = 1;

if ($opt{D}) {
	&misc::Daemonize;
}
my $maxlen	= 512;
my $port	= ($opt{p})?$opt{p}:514;
my $devup	= 0;

my $sock = IO::Socket::INET->new(LocalPort => $port, Proto => 'udp') or die "socket: $@";
print "Awaiting syslog messages on port $port\n" if $opt{v};
while ($sock->recv(my $msg, $maxlen)) {
		my $now = time;
		if($now - $misc::rrdstep > $devup){							# refresh devices if older than a discovery...
			$devup = $now;
			&db::ReadDev();
			undef (%dip);
			foreach my $d (keys %dev){
				$dip{$dev{$d}{'ip'}} = $d;
			}
			print "Devs updated!\n" if $opt{v};
		}
	my($client_port, $client_ip) = sockaddr_in($sock->peername);
	my $ip = inet_ntoa($client_ip);
# I'll put some aggregation here, when I find the time...
	&Process($ip,$msg);
}
die "recv: $!";

#===================================================================
# Display some help
# Parameters:	-
# Global:		-
# Return:		-
#===================================================================
sub HELP_MESSAGE {
	print "\n";
	print "usage: syslog.pl <Option(s)>\n\n";
	print "---------------------------------------------------------------------------\n";
	print "Options:\n";
	print "-D		daemonize moni.pl\n";
	print "-v		verbose output\n\n";
	print "syslog $main::VERSION (C) 2001-2009 Remo Rickli (and contributors)\n\n";
	die;
}

#===================================================================
# Process Message
# Parameters:	source IP, message
# Global:		-
# Return:		dies upon DB error!
#===================================================================
sub Process {

	my $src = $_[0];
	my $now = time;

	my $level = 10;
	if( exists $dip{$_[0]}){
		$src = $dip{$_[0]};
		$level = 50;
	}
	my $pri = $_[1];
	print "$src $pri\n"  if $opt{v};
	$pri =~ s/<(\d+)>.*/$1/;
	my $sev = ($pri & 7);
	if($level == 50){
		   if ($sev == 4){$level = 100}
		elsif ($sev == 3){$level = 150}
		elsif ($sev =~ /[012]/){$level = 200}
	}
	my $msg = $_[1];
	$msg =~ s/<(\d+)>(.*)/$2/;
	$msg =~ s/[^\w\t\/\Q(){}[]!@#$%^&*-+=',.:<>? \E]//g;
	$msg = substr($msg,0,255);
	print "$src ($_[0])\tS:$sev\tL:$level\t$msg\n"  if $opt{v};
	if( $level >= $misc::msglevel or $dev{$src}{mi} ne "" and $msg =~ /$dev{$src}{mi}/){
		#&mon::SendSMS($msg);
		&mon::SendMail("Syslog Alert","$src: $msg!");
	}
	if( ! &db::Insert('messages','level,time,source,info',"\"$level\",\"$now\",\"$src\",\"$msg\"") ){
		die "DB error messages!\n";
	}
}
